'''
A module to control the LArPix chip.

'''

from __future__ import absolute_import
from __future__ import print_function
import time
import json
import os
import errno
import math
import warnings
import struct
import sys
from collections import OrderedDict

from bitarray import bitarray

from . import bitarrayhelper as bah

from .key import *
from .chip import *
from .configuration import *
from .configuration import _Smart_List
from .controller import *
from .packet import *

from .logger import Logger
from . import configs



from bitarray import bitarray

from .. import bitarrayhelper as bah
from ..key import Key

class Packet_v2(object):
    '''
    Representation of a 64 bit LArPix v2 UART data packet.

    Packet_v2 objects are internally represented as bitarrays, but a variety of
    helper properties allow one to access and set the data stored in the bitarrays in
    a natural fashion. E.g.::

        p = Packet_v2() # initialize a packet of zeros
        p.packet_type # fetch packet type bits and convert to uint
        p.packet_type = 2 # set the packet type to a config write packet
        print(p.bits) # the bits have been updated!

    Packet_v2 objects don't enforce any value validation, so set these fields with caution!

    In FIFO diagnostics mode, the bits are to be interpreted in a different way.
    At this point, there is no way for the ``Packet_v2`` to automatically know if it
    is in FIFO diagnostics mode. If you are operating the chips in this mode you
    can manually set ``Packet_v2.fifo_diagnostics_enabled = True`` to interpret
    data packets in this fashion by default, or set ``packet.fifo_diagnostics_enabled = True``
    for a single packet. Just remember that if you modify ``packet.fifo_diagnostics_enabled``,
    it will no longer use the default.

    '''

    asic_version = 2
    size = 64
    num_bytes = 8

    # shared by all packet types
    packet_type_bits = slice(0,2)
    chip_id_bits = slice(2,10)
    downstream_marker_bits = slice(62,63)
    parity_bits = slice(63,64)
    parity_calc_bits = slice(0,63)

    # only data packets
    channel_id_bits = slice(10,16)
    timestamp_bits = slice(16,47)
    first_packet_bits = slice(47,48)
    dataword_bits = slice(48,56)
    trigger_type_bits = slice(56,58)
    local_fifo_bits = slice(58,60)
    shared_fifo_bits = slice(60,62)
    # only if fifo diagnostics enabled
    fifo_diagnostics_timestamp_bits = slice(16,32)
    local_fifo_events_bits = slice(44,46)
    shared_fifo_events_bits = slice(32,44)

    # only read/write packets
    register_address_bits = slice(10,18)
    register_data_bits = slice(18,26)

    fifo_diagnostics_enabled = False

    DATA_PACKET = 0
    TEST_PACKET = 1
    CONFIG_WRITE_PACKET = 2
    CONFIG_READ_PACKET = 3

    NORMAL_TRIG = 0
    EXT_TRIG = 1
    CROSS_TRIG = 2
    PERIODIC_TRIG = 3

    endian = 'little'

    def __init__(self, bytestream=None):
        if bytestream is None:
            self.bits = bitarray(self.size,endian=self.endian)
            self.bits.setall(False)
            return
        elif len(bytestream) == self.num_bytes:
            self.bits = bitarray(endian=self.endian)
            self.bits.frombytes(bytestream)
        else:
            raise ValueError('Invalid number of bytes: %s' %
                    len(bytestream))

    def __eq__(self, other):
        return self.bits == other.bits

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        strings = []
        if hasattr(self, 'direction'):
            strings += ['Direction: {}'.format(self.direction)]
        strings += ['Key: {}'.format(self.chip_key)]
        strings += ['Chip: {}'.format(self.chip_id)]
        if self.downstream_marker:
            strings += ['Downstream']
        else:
            strings += ['Upstream']
        if self.packet_type == self.DATA_PACKET:
            strings += ['Data']
            strings += ['Channel: {}'.format(self.channel_id)]
            strings += ['Timestamp: {}'.format(self.timestamp)]
            strings += ['First packet: {}'.format(self.first_packet)]
            strings += ['Dataword: {}'.format(self.dataword)]
            strings += ['Trigger: {}'.format({
                self.NORMAL_TRIG: 'normal',
                self.EXT_TRIG: 'external',
                self.CROSS_TRIG: 'cross',
                self.PERIODIC_TRIG: 'periodic'
                }[self.trigger_type])]
            if self.local_fifo_full:
                strings += ['Local FIFO 100%']
            elif self.local_fifo_half:
                strings += ['Local FIFO >50%']
            else:
                strings += ['Local FIFO ok']
            if self.shared_fifo_full:
                strings += ['Shared FIFO 100%']
            elif self.shared_fifo_half:
                strings += ['Shared FIFO >50%']
            else:
                strings += ['Shared FIFO ok']
            if self.fifo_diagnostics_enabled:
                strings += ['Local FIFO: {}'.format(self.local_fifo_events)]
                strings += ['Shared FIFO: {}'.format(self.shared_fifo_events)]
        elif self.packet_type == self.TEST_PACKET:
            strings += ['Test']
        elif self.packet_type in (self.CONFIG_READ_PACKET, self.CONFIG_WRITE_PACKET):
            strings += [{
                self.CONFIG_READ_PACKET: 'Read',
                self.CONFIG_WRITE_PACKET: 'Write'
            }[self.packet_type]]
            strings += ['Register: {}'.format(self.register_address)]
            strings += ['Value: {}'.format(self.register_data)]

        if hasattr(self, 'receipt_timestamp'):
            strings += ['Receipt TS: {}'.format(self.receipt_timestamp)]
        strings += ['Parity: {} (valid: {})'.format(self.parity,
            self.has_valid_parity())]
        return '[ ' + ' | '.join(strings) + ' ]'

    def __repr__(self):
        return 'Packet_v2(' + str(self.bytes()) + ')'

    def bytes(self):
        '''
        Create bytes that represent the packet.

        Byte 0 is still the first byte to send out and contains bits [0:7]

        '''
        return self.bits.tobytes() #[::-1]

    def export(self):
        '''
        Return a dict representation of the packet

        '''
        type_map = {
                self.TEST_PACKET: 'test',
                self.DATA_PACKET: 'data',
                self.CONFIG_WRITE_PACKET: 'config write',
                self.CONFIG_READ_PACKET: 'config read'
                }
        d = {}
        d['asic_version'] = self.asic_version
        d['chip_key'] = str(self.chip_key) if self.chip_key else None
        d['io_group'] = self.io_group
        d['io_channel'] = self.io_channel
        d['bits'] = self.bits.to01()
        d['type_str'] = type_map[self.packet_type]
        d['packet_type'] = self.packet_type
        d['chip_id'] = self.chip_id
        d['downstream_marker'] = self.downstream_marker
        d['parity'] = self.parity
        d['valid_parity'] = self.has_valid_parity()
        ptype = self.packet_type
        if ptype == self.TEST_PACKET:
            pass
        elif ptype == self.DATA_PACKET:
            d['channel_id'] = self.channel_id
            d['timestamp'] = self.timestamp
            d['first_packet'] = self.first_packet
            d['dataword'] = self.dataword
            d['trigger_type'] = self.trigger_type
            d['local_fifo'] = self.local_fifo
            d['shared_fifo'] = self.shared_fifo
            if self.fifo_diagnostics_enabled:
                d['local_fifo_events'] = self.local_fifo_events
                d['shared_fifo_events'] = self.shared_fifo_events
        elif (ptype == self.CONFIG_READ_PACKET or ptype ==
                self.CONFIG_WRITE_PACKET):
            d['register_address'] = self.register_address
            d['register_data'] = self.register_data
        if hasattr(self,'direction'):
            d['direction'] = self.direction
        if hasattr(self,'receipt_timestamp'):
            d['receipt_timestamp'] = self.receipt_timestamp
        return d

    def from_dict(self, d):
        '''
        Load from a dict of values, inverse of ``Packet_v2.export``

        .. note:: If there is a disagreement between the bits specified and any of the values, this method has undefined results.

        '''
        if not d['asic_version'] == self.asic_version:
            raise ValueError('invalid asic version {}'.format(d['asic_version']))
        if 'type' in d and d['type'] not in (self.DATA_PACKET, self.TEST_PACKET, self.CONFIG_WRITE_PACKET, self.CONFIG_READ_PACKET):
            raise ValueError('invalid packet type for Packet_v2')
        if 'local_fifo_events' in d or 'shared_fifo_events' in d:
            self.fifo_diagnostics_enabled = True
        for key, value in d.items():
            if key in ('type_str', 'valid_parity'):
                continue
            elif key == 'bits':
                self.bits = bitarray(value)
            else:
                setattr(self, key, value)

    @property
    def chip_key(self):
        ''''''
        if hasattr(self, '_chip_key'):
            return self._chip_key
        if self.io_group is None or self.io_channel is None:
            return None
        self._chip_key = Key(self.io_group, self.io_channel, self.chip_id)
        return self._chip_key

    @chip_key.setter
    def chip_key(self, value):
        # remove cached key
        if hasattr(self, '_chip_key'):
            del self._chip_key
        if value is None:
            self.io_channel = None
            self.io_group = None
            return
        if isinstance(value, Key):
            self.io_group = value.io_group
            self.io_channel = value.io_channel
            self.chip_id = value.chip_id
            return
        # try again by casting as a Key
        self.chip_key = Key(value)

    @property
    def io_group(self):
        ''''''
        if hasattr(self, '_io_group'):
            return self._io_group
        return None

    @io_group.setter
    def io_group(self, value):
        if hasattr(self, '_chip_key'):
            # remove cached key
            del self._chip_key
        if value is None:
            if hasattr(self, '_io_group'):
                del self._io_group
            return
        # no value validation!
        self._io_group = value

    @property
    def io_channel(self):
        if hasattr(self, '_io_channel'):
            return self._io_channel
        return None

    @io_channel.setter
    def io_channel(self, value):
        if hasattr(self, '_chip_key'):
            # remove cached key
            del self._chip_key
        if value is None:
            if hasattr(self, '_io_channel'):
                del self._io_channel
            return
        # no value validation!
        self._io_channel = value

    @property
    def timestamp(self):
        if self.fifo_diagnostics_enabled:
            return bah.touint(self.bits[self.fifo_diagnostics_timestamp_bits], endian=self.endian)
        return bah.touint(self.bits[self.timestamp_bits], endian=self.endian)

    @timestamp.setter
    def timestamp(self, value):
        if self.fifo_diagnostics_enabled:
            self.bits[self.fifo_diagnostics_timestamp_bits] = bah.fromuint(value, self.fifo_diagnostics_timestamp_bits, endian=self.endian)
        else:
            self.bits[self.timestamp_bits] = bah.fromuint(value, self.timestamp_bits, endian=self.endian)

    @property
    def local_fifo_half(self):
        return self.local_fifo%2

    @local_fifo_half.setter
    def local_fifo_half(self, value):
        self.local_fifo = self.local_fifo_full*2 + value

    @property
    def local_fifo_full(self):
        return self.local_fifo//2

    @local_fifo_full.setter
    def local_fifo_full(self, value):
        self.local_fifo = value*2 + self.local_fifo_half

    @property
    def shared_fifo_half(self):
        return self.shared_fifo%2

    @shared_fifo_half.setter
    def shared_fifo_half(self, value):
        self.shared_fifo = self.shared_fifo_full*2 + value

    @property
    def shared_fifo_full(self):
        return self.shared_fifo//2

    @shared_fifo_full.setter
    def shared_fifo_full(self, value):
        self.shared_fifo = value*2 + self.shared_fifo_half

    def compute_parity(self):
        return 1 - (self.bits[self.parity_calc_bits].count(True) % 2)

    def assign_parity(self):
        self.parity = self.compute_parity()

    def has_valid_parity(self):
        return self.parity == self.compute_parity()

    @property
    def local_fifo_events(self):
        if self.fifo_diagnostics_enabled:
            bit_slice = self.local_fifo_events_bits
            return bah.touint(self.bits[bit_slice], endian=self.endian)
        return None

    @local_fifo_events.setter
    def local_fifo_events(self, value):
        if self.fifo_diagnostics_enabled:
            bit_slice = self.local_fifo_events_bits
            self.bits[bit_slice] = bah.fromuint(value, bit_slice, endian=self.endian)

    @property
    def shared_fifo_events(self):
        if self.fifo_diagnostics_enabled:
            bit_slice = self.shared_fifo_events_bits
            return bah.touint(self.bits[bit_slice], endian=self.endian)
        return None

    @shared_fifo_events.setter
    def shared_fifo_events(self, value):
        if self.fifo_diagnostics_enabled:
            bit_slice = self.shared_fifo_events_bits
            self.bits[bit_slice] = bah.fromuint(value, bit_slice, endian=self.endian)

    @property
    def chip_id(self):
        bit_slice = self.chip_id_bits
        return bah.touint(self.bits[bit_slice], endian=self.endian)

    @chip_id.setter
    def chip_id(self, value):
        if hasattr(self,'_chip_key'):
            del self._chip_key
        bit_slice = self.chip_id_bits
        self.bits[bit_slice] = bah.fromuint(value, bit_slice, endian=self.endian)

    def _basic_getter(name):
        def basic_getter_func(self):
            bit_slice = getattr(self, name + '_bits')
            return bah.touint(self.bits[bit_slice], endian=self.endian)
        return basic_getter_func

    def _basic_setter(name):
        def basic_setter_func(self, value):
            bit_slice = getattr(self, name + '_bits')
            self.bits[bit_slice] = bah.fromuint(value, bit_slice, endian=self.endian)
        return basic_setter_func

    packet_type = property(_basic_getter('packet_type'),_basic_setter('packet_type'))
    downstream_marker = property(_basic_getter('downstream_marker'),_basic_setter('downstream_marker'))
    parity = property(_basic_getter('parity'),_basic_setter('parity'))
    channel_id = property(_basic_getter('channel_id'),_basic_setter('channel_id'))
    dataword = property(_basic_getter('dataword'),_basic_setter('dataword'))
    first_packet = property(_basic_getter('first_packet'),_basic_setter('first_packet'))
    trigger_type = property(_basic_getter('trigger_type'),_basic_setter('trigger_type'))
    register_address = property(_basic_getter('register_address'),_basic_setter('register_address'))
    register_data = property(_basic_getter('register_data'),_basic_setter('register_data'))
    local_fifo = property(_basic_getter('local_fifo'),_basic_setter('local_fifo'))
    shared_fifo = property(_basic_getter('shared_fifo'),_basic_setter('shared_fifo'))



import time
import zmq
import sys
import copy
import warnings

from larpix.io import MultiZMQ_IO
from larpix import Packet, Key

class ZMQ_IO(MultiZMQ_IO):
    '''
    The ZMQ_IO object interfaces with the Bern LArPix v2 module using
    the ZeroMQ communications protocol. This class wraps the
    ``io.multizmq_io.MultiZMQ_IO`` class, and enables a slightly simpler chip
    key formatting in the special case that you are only interfacing with a
    single daq board.

    This object handles the required communications, and also has extra
    methods for additional functionality, including system reset, packet
    count, clock frequency, and more.

    '''
    _valid_config_classes = MultiZMQ_IO._valid_config_classes + ['ZMQ_IO']

    def __init__(self, config_filepath=None, **kwargs):
        super(ZMQ_IO, self).__init__(config_filepath=config_filepath, **kwargs)
        self._address = list(self._io_group_table.values())[0]

    def load(self, filepath=None):
        super(ZMQ_IO, self).load(filepath)
        if len(self._io_group_table.inv) != 1:
            raise RuntimeError('multiple adresses found in configuration - use '
                'MultiZMQ_IO if you\'d like to connect to multiple systems')

    @property
    def sender(self):
        return self.senders[self._address]
    @sender.setter
    def sender(self, val):
        self.senders[self._address] = val

    @property
    def receiver(self):
        return self.receivers[self._address]
    @receiver.setter
    def receiver(self, val):
        self.receivers[self._address] = val

    @property
    def sender_replies(self):
        return super(ZMQ_IO, self).sender_replies[self._address]
    @sender_replies.setter
    def sender_replies(self, val):
        super(ZMQ_IO, self).sender_replies[self._address] = val

    def decode(self, msgs, **kwargs):
        '''
        Convert a list ZMQ messages into packets

        '''
        if not 'address' in kwargs:
            return super(ZMQ_IO, self).decode(msgs, address=self._address, **kwargs)
        return super(ZMQ_IO, self).decode(msgs, **kwargs)

    def reset(self):
        '''
        Send a reset pulse to the LArPix ASICs.

        '''
        return super(ZMQ_IO, self).reset(addresses=[self._address])[self._address]

    def set_clock(self, freq_khz):
        '''
        Set the LArPix CLK2X freqency (in kHz).

        :param freq_khz: CLK2X freq in khz to set

        '''
        return super(ZMQ_IO, self).set_clock(freq_khz=freq_khz, addresses=[self._address])[self._address]

    def set_testpulse_freq(self, divisor):
        '''
        Set the testpulse frequency, computed by dividing the CLK2X
        frequency by ``divisor``.

        '''
        return super(ZMQ_IO, self).set_testpulse_freq(divisor=divisor, address=self._address)

    def get_packet_count(self, io_channel):
        '''
        Get the number of packets received, as determined by the number
        of UART "start" bits processed.

        '''
        return super(ZMQ_IO, self).get_packet_count(io_channel=io_channel, address=self._address)

    def ping(self):
        '''
        Send a ping to the system and return True if the first two bytes
        of the response are b'OK'.

        '''
        return super(ZMQ_IO, self).ping(addresses=[self._address])[self._address]



import pytest
import zmq

@pytest.fixture
def temp_logfilename():
    return 'test.h5'

def test_min_example(capsys):
    from larpix import Controller, Packet_v2
    from larpix.io import FakeIO
    from larpix.logger import StdoutLogger
    controller = Controller()
    controller.io = FakeIO()
    controller.logger = StdoutLogger(buffer_length=0)
    controller.logger.enable()
    chip1 = controller.add_chip('1-1-2', version=2)  # (access key)
    chip1.config.threshold_global = 25
    controller.write_configuration('1-1-2', chip1.config.register_map['threshold_global']) # chip key, register 64
    assert capsys.readouterr().out == '[ Key: 1-1-2 | Chip: 2 | Upstream | Write | Register: 64 | Value: 25 | Parity: 1 (valid: True) ]\nRecord: [ Key: 1-1-2 | Chip: 2 | Upstream | Write | Register: 64 | Value: 25 | Parity: 1 (valid: True) ]\n'
    packet = Packet_v2(b'\x08\x14\x15\xcd[\x07\x91@')
    packet_bytes = packet.bytes()
    pretend_input = ([packet], packet_bytes)
    controller.io.queue.append(pretend_input)
    controller.run(0.05, 'test run')
    assert capsys.readouterr().out == 'Record: [ Key: None | Chip: 2 | Downstream | Data | Channel: 5 | Timestamp: 123456789 | First packet: 0 | Dataword: 145 | Trigger: normal | Local FIFO ok | Shared FIFO ok | Parity: 0 (valid: True) ]\n'
    print(controller.reads[0])
    assert capsys.readouterr().out == '[ Key: None | Chip: 2 | Downstream | Data | Channel: 5 | Timestamp: 123456789 | First packet: 0 | Dataword: 145 | Trigger: normal | Local FIFO ok | Shared FIFO ok | Parity: 0 (valid: True) ]\n'


def test_tutorial(capsys, tmpdir, temp_logfilename):
    from larpix import Controller, Packet_v2

    from larpix.io import FakeIO
    from larpix.logger import StdoutLogger
    controller = Controller()
    controller.io = FakeIO()
    controller.logger = StdoutLogger(buffer_length=0)
    controller.logger.enable()

    chipid = 5
    chip_key = '1-1-5'
    chip5 = controller.add_chip(chip_key, version=2)
    chip5 = controller[chip_key]
    chip5 = controller[1,1,5]


    from larpix import Key
    example_key = Key(1,2,3)


    assert example_key.io_group  == 1
    assert example_key.io_channel  == 2
    assert example_key.chip_id  == 3
    example_key.to_dict()


    controller.load('controller/v2_example.json')
    print(controller.chips) # chips that have been loaded into controller
    list(controller.network[1][1]['miso_ds'].edges) # all links contained in the miso_ds graph
    list(controller.network[1][1]['miso_us'].nodes) # all nodes within the miso_us graph
    list(controller.network[1][1]['mosi'].edges) # all links within the mosi graph

    list(controller.network[1][1]['mosi'].in_edges(2)) # all links pointing to chip 2 in mosi graph
    list(controller.network[1][1]['miso_ds'].successors(3)) # all chips receiving downstream data packets from chip 3
    controller.network[1][1]['mosi'].edges[(3,2)]['uart'] # check the physical uart channel that chip 2 listens to chip 3 via
    controller.network[1][1]['mosi'].nodes[2]['root'] # check if designated root chip

    controller.init_network(1,1) # issues packets required to initialize the 1,1 hydra network
    print(controller['1-1-2'].config.chip_id)
    print(controller['1-1-3'].config.enable_miso_downstream)

    controller.reset_network(1,1)

    controller.init_network(1,1,2) # configures only chip 2
    controller.init_network(1,1,3) # configures only chip 3

    assert isinstance(controller.get_network_keys(1,1), list) # gets a list of chip keys starting at the root node and descending
    assert isinstance(controller.get_network_keys(1,1,root_first_traversal=False), list) # get list of chip keys starting at deepest chips and ascending


    chip5.config.threshold_global = 35  # entire register = 1 number
    chip5.config.enable_periodic_reset = 1  # one bit as part of a register
    chip5.config.channel_mask[20] = 1  # one bit per channel


    controller.write_configuration(chip_key)  # send all registers
    controller.write_configuration(chip_key, 32)  # send only register 32
    controller.write_configuration(chip_key, [32, 50])  # send registers 32 and 50


    threshold_global_reg = chip5.config.register_map['threshold_global']


    threshold_global_name = chip5.config.register_map_inv[64]


    packets = chip5.get_configuration_read_packets()
    bytestream = b'bytes for the config read packets'
    controller.io.queue.append((packets, bytestream))

    controller.read_configuration(chip_key)

    packets = [Packet_v2()] * 40
    bytestream = b'bytes from the first set of packets'
    controller.io.queue.append((packets, bytestream))
    packets2 = [Packet_v2()] * 30
    bytestream2 = b'bytes from the second set of packets'
    controller.io.queue.append((packets2, bytestream2))

    controller.start_listening()
    # Data arrives...
    packets, bytestream = controller.read()
    # More data arrives...
    packets2, bytestream2 = controller.read()
    controller.stop_listening()
    message = 'First data arrived!'
    message2 = 'More data arrived!'
    controller.store_packets(packets, bytestream, message)
    controller.store_packets(packets, bytestream2, message2)


    packets = [Packet_v2()] * 5
    bytestream = b'[bytes from read #%d] '
    for i in range(100):
        controller.io.queue.append((packets, bytestream%i))

    duration = 0.1  # seconds
    message = '10-second data run'
    controller.run(duration, message)


    run1 = controller.reads[0]
    first_packet = run1[0]  # Packet object
    first_ten_packets = run1[0:10]  # smaller PacketCollection object

    first_packet_bits = run1[0, 'bits']  # string representation of bits in packet
    first_ten_packet_bits = run1[0:10, 'bits']  # list of strings


    print(run1)  # prints the contents of the packets
    print(run1[10:30])  # prints 20 packets from the middle of the run


    packet = run1[0]
    # all packets
    packet.packet_type  # unique in that it gives the bits representation
    packet.chip_id  # all other properties return Python numbers
    packet.chip_key # key for association to a unique chip (can be None)
    packet.parity
    packet.downstream_marker

    # data packets
    packet.channel_id
    packet.dataword
    packet.timestamp
    packet.trigger_type
    packet.local_fifo
    packet.shared_fifo
    # config packets
    packet.register_address
    packet.register_data

    # if fifo_diagnostics enabled on a given chip the timestamp of data packets
    # are interpreted differently (note that this is not done automatically and will
    # need to be performed manually on each packet)
    packet.fifo_diagnostics_enabled = True
    packet.timestamp
    packet.local_fifo_events
    packet.shared_fifo_events


    from larpix.logger import HDF5Logger
    controller.logger = HDF5Logger(filename=temp_logfilename, directory=str(tmpdir), buffer_length=10000) # a filename of None uses the default filename formatting
    controller.logger.enable() # opens hdf5 file and starts tracking all communications

    controller.logger = HDF5Logger(filename=temp_logfilename,
            directory=str(tmpdir), enabled=True)

    controller.verify_configuration()
    controller.logger.flush()


    controller.logger.disable() # stop tracking
    # any communication here is ignored
    controller.logger.enable() # start tracking again
    controller.logger.is_enabled() # returns True if tracking


    controller.logger.disable()


    import h5py
    datafile = h5py.File(tmpdir.join(temp_logfilename))


    assert '_header' in datafile.keys()
    assert 'packets' in datafile.keys()
    assert 'messages' in datafile.keys()
    assert list(datafile['_header'].attrs) == ['created', 'modified', 'version']


    raw_value = datafile['packets'][0] # e.g. (b'0-246', 3, 246, 1, 1, -1, -1, -1, -1, -1, -1, 0, 0)
    raw_values = datafile['packets'][-100:] # last 100 packets in file


    packet_repr = raw_values[0:1]
    packet_repr['chip_id'] # chip id for packet, e.g. 246
    packet_repr['dataword'] # list of ADC values for each packet
    packet_repr.dtype # description of data type (with names of each column)

    # all packets' ADC counts, including non-data packets
    raw_values['dataword']
    # Select based on data type using a numpy bool / "mask" array:
    raw_values['dataword'][raw_values['packet_type'] == 0] # all data packets' ADC counts

    datafile.close()

def test_running_with_bern_daq_v1():
    from larpix import Controller
    from larpix.io import ZMQ_IO

    controller = Controller()
    controller.io = ZMQ_IO(config_filepath='io/loopback.json')

    controller.load('controller/pcb-2_chip_info.json')

    for key,chip in controller.chips.items():
        chip.config.load('chip/quiet.json')
        print(key, chip.config)

def test_running_with_pacman_v1r1():
    from larpix import Controller
    from larpix.io import PACMAN_IO
    controller = Controller()
    controller.io = PACMAN_IO(config_filepath='io/pacman.json',timeout=1000)
    controller.load('controller/network-3x3-tile-channel0.json')
    assert isinstance(controller.io.ping(),dict)

    with pytest.raises(zmq.ZMQError):
        controller.io.set_vddd()
    with pytest.raises(zmq.ZMQError):
        controller.io.set_vdda()

    # First bring up the network using as few packets as possible
    controller.io.group_packets_by_io_group = False # this throttles the data rate to avoid FIFO collisions
    for io_group, io_channels in controller.network.items():
        for io_channel in io_channels:
            with pytest.raises(zmq.ZMQError):
                controller.init_network(io_group, io_channel)

    # Configure the IO for a slower UART and differential signaling
    controller.io.double_send_packets = True # double up packets to avoid 512 bug when configuring
    for io_group, io_channels in controller.network.items():
        for io_channel in io_channels:
            chip_keys = controller.get_network_keys(io_group,io_channel,root_first_traversal=False)
            for chip_key in chip_keys:
                controller[chip_key].config.clk_ctrl = 1
                controller[chip_key].config.enable_miso_differential = [1,1,1,1]
                with pytest.raises(zmq.ZMQError):
                    controller.write_configuration(chip_key, 'enable_miso_differential')
                    controller.write_configuration(chip_key, 'clk_ctrl')
    for io_group, io_channels in controller.network.items():
        for io_channel in io_channels:
            with pytest.raises(zmq.ZMQError):
                controller.io.set_uart_clock_ratio(io_channel, 4, io_group=io_group)

    controller.io.double_send_packets = False
    controller.io.group_packets_by_io_group = True



'''
A module with convenience functions for bitarray.

'''
from bitarray import bitarray

def fromuint(val, nbits, endian='big'):
    try:
        if isinstance(nbits, slice):
            nbits = abs(nbits.stop - nbits.start)
        if endian[0] == 'b':
            string = bin(val)[2:].zfill(nbits)
            return bitarray(string)
        string = bin(val)[-1:1:-1].ljust(nbits,'0')
        return bitarray(string)
    except TypeError:
        return val

def touint(bits, endian='big'):
    bin_string = bits.to01()
    if endian[0] == 'b':
        return int(bin_string, 2)
    return int(bin_string[::-1], 2)



import time
import zmq
import sys
from collections import defaultdict
import warnings
import bidict

from larpix.io import IO
from larpix import Packet, Key, Packet_v2, Packet_v1
from larpix.configs import load
from larpix.format.message_format import dataserver_message_decode

class MultiZMQ_IO(IO):
    '''
    The MultiZMQ_IO object interfaces with a network of Bern LArPix v2 modules
    using Igor's ZeroMQ communications protocol.

    This object handles the required communications, and also has extra
    methods for additional functionality, including system reset, packet
    count, clock frequency, and more.

    By default, when creating a MultiZMQ_IO object, the ``io/default.json``
    configuration will attempt to be loaded unless otherwise specified. The path relative to the pwd is checked first,
    followed by the path of the larpix-control installation.

    '''
    _valid_config_classes = ['MultiZMQ_IO']

    def __init__(self, config_filepath=None, miso_map=None, mosi_map=None):
        super(MultiZMQ_IO, self).__init__()
        self.load(config_filepath)

        if miso_map is None:
            self._miso_map = {}
        else:
            self._miso_map = miso_map
        if mosi_map is None:
            self._mosi_map = {}
        else:
            self._mosi_map = mosi_map

        self.context = zmq.Context()
        self.senders = bidict.bidict()
        self.receivers = bidict.bidict()
        for address in self._io_group_table.inv:
            self.senders[address] = self.context.socket(zmq.REQ)
            self.receivers[address] = self.context.socket(zmq.SUB)
        self.hwm = 20000
        for receiver in self.receivers.values():
            receiver.set_hwm(self.hwm)
        for address in self._io_group_table.inv:
            send_address = 'tcp://' + address + ':5555'
            receive_address = 'tcp://' + address + ':5556'
            self.senders[address].connect(send_address)
            self.receivers[address].connect(receive_address)
        self._sender_replies = defaultdict(list)
        self.poller = zmq.Poller()
        for receiver in self.receivers.values():
            self.poller.register(receiver, zmq.POLLIN)

    @property
    def sender_replies(self):
        return self._sender_replies
    @sender_replies.setter
    def sender_replies(self, val):
        self._sender_replies = val

    def send(self, packets):
        self.sender_replies = defaultdict(list)
        send_time = time.time()
        addresses = [self._io_group_table[packet.io_group] for packet in packets]
        msg_datas = self.encode(packets)
        for address, msg_data in zip(addresses, msg_datas):
            tosend = b'SNDWORD ' + msg_data
            self.senders[address].send(tosend)
            self.sender_replies[address].append(self.senders[address].recv())

    def start_listening(self):
        if self.is_listening:
            raise RuntimeError('Already listening')
        super(MultiZMQ_IO, self).start_listening()
        for receiver in self.receivers.values():
            receiver.setsockopt(zmq.SUBSCRIBE, b'')

    def stop_listening(self):
        if not self.is_listening:
            raise RuntimeError('Already not listening')
        super(MultiZMQ_IO, self).stop_listening()
        for receiver in self.receivers.values():
            receiver.setsockopt(zmq.UNSUBSCRIBE, b'')

    def decode(self, msgs, address, **kwargs):
        '''
        Convert a list ZMQ messages into packets

        '''
        return_packets = dataserver_message_decode(msgs, version=(1,0), io_group=self._io_group_table.inv[address], **kwargs)
        if self._miso_map.keys():
            for packet in return_packets:
                if hasattr(packet, 'io_channel') and packet.io_channel in self._miso_map.keys():
                    packet.io_channel = self._miso_map[packet.io_channel]
        return return_packets

    def encode(self, packets):
        '''
        Encode a list of packets into ZMQ messages
        '''
        msg_data = []
        for packet in packets:
            io_chain = packet.io_channel
            if io_chain in self._mosi_map.keys():
                io_chain = self._mosi_map[io_chain]
            fmt_bytestr = b'0x%s %d'
            if isinstance(packet, Packet_v1):
                fmt_bytestr = b'0x00%s %d'
            if sys.version_info[0] < 3:
                msg_data += [fmt_bytestr % (packet.bytes()[::-1].encode('hex'), io_chain)]
            else:
                msg_data += [fmt_bytestr % (packet.bytes()[::-1].hex().encode(), io_chain)]
        print(msg_data)
        return msg_data

    def empty_queue(self):
        '''
        Process any awaiting messages from all ZMQ connections. Will continue
        to read until the `hwm` is reached or there are no more awaiting messages.

        :returns: 2-`tuple` containing a list of received packets and the full bytestream
        '''
        packets = []
        address_list = []
        bytestream_list = []
        bytestream = b''
        n_recv = 0
        read_time = time.time()
        message = ''
        while self.poller.poll(0) and n_recv < self.hwm:
            events = dict(self.poller.poll(0))
            for socket, n_events in events.items():
                for _ in range(n_events):
                    message = socket.recv()
                    n_recv += 1
                    bytestream_list += [message]
                    address_list += [self.receivers.inv[socket]]
        for message, address in zip(bytestream_list, address_list):
            #print(message)
            packets += self.decode([message], address=address)
        #print('len(bytestream_list) = %d' % len(bytestream_list))
        bytestream = b''.join(bytestream_list)
        return packets, bytestream

    def reset(self, addresses=None):
        '''
        Send a reset pulse to the LArPix ASICs.

        :param addresses: ``list`` of daq board addresses to reset, if ``None`` reset all addresses

        '''
        if addresses is None:
            return self.reset(addresses=self.senders.keys())

        for address in addresses:
            self.senders[address].send(b'SYRESET')
        return_dict = {}
        for address in addresses:
            return_dict[address] = self.senders[address].recv()
        return return_dict

    def set_clock(self, freq_khz, addresses=None):
        '''
        Set the LArPix CLK2X freqency (in kHz).

        :param freq_khz: CLK2X freq in khz to set

        :param addresses: ``list`` of daq board addresses to change frequency, if ``None`` modifies all addresses

        '''
        if addresses is None:
            return self.set_clock(freq_khz, addresses=self.senders.keys())

        for address in addresses:
            self.senders[address].send(b'SETFREQ %s' % hex(freq_khz).encode())
        return_dict = {}
        for address in addresses:
            return_dict[address] = self.senders[address].recv()
        return return_dict

    def set_testpulse_freq(self, divisor, address):
        '''
        Set the testpulse frequency, computed by dividing the CLK2X
        frequency by ``divisor``.

        :param divisor: test pulse frequency divisor

        :param address: daq board addresses to change test pulse freq

        '''
        self.senders[address].send(b'SETFTST %s' % hex(divisor).encode())
        return self.senders[address].recv()

    def get_packet_count(self, io_channel, address):
        '''
        Get the number of packets received, as determined by the number
        of UART "start" bits processed.

        :param io_channel: IO channel to check

        :param address: address of daq board

        '''
        self.senders[address].send(b'GETSTAT %d' % io_channel)
        result = self.senders[address].recv()
        space = result.find(b' ')
        number = int(result[:space])
        return number

    def ping(self, addresses=None):
        '''
        Send a ping to the system.

        :param addresses: ``list`` of daq board addresses to ping, if ``None`` ping all addresses

        :returns: ``dict`` with one entry per address. Value is ``True`` if first two bytes of the response are b'OK'.

        '''
        if addresses is None:
            return self.ping(addresses=self.senders.keys())

        for address in addresses:
            self.senders[address].send(b'PING_HB')
        result = {}
        for address in addresses:
            received_msg = self.senders[address].recv()
            result[address] = received_msg[:2] == b'OK'
        return result

    def cleanup(self):
        '''
        Close the ZMQ objects to prevent a memory leak.

        This method is only required if you plan on instantiating a new
        ``MultiZMQ_IO`` object.

        '''
        for address in self.senders.keys():
            self.senders[address].close(linger=0)
            self.receivers[address].close(linger=0)
        self.context.term()



from bitarray import bitarray
import struct

from .. import bitarrayhelper as bah
from ..key import Key

class TimestampPacket(object):
    '''
    A packet-like object which just contains an integer timestamp.

    This class implements many methods used by Packet, so it functions
    smoothly in lists of packets and in PacketCollection.

    If neither ``timestamp`` nor ``code`` is provided then this
    TimestampPacket will have a timestamp of ``None`` until it is
    manually set.

    :param timestamp: optional, integer timestamp of this packet
    :param code: optional, encoded timestamp as a 7-byte unsigned int
        obtainable from calling the ``bytes`` method.

    '''
    size = 56
    chip_key = None
    def __init__(self, timestamp=None, code=None):
        self.packet_type = 4
        if code:
            self.timestamp = struct.unpack('<Q', code + b'\x00')[0]
        else:
            self.timestamp = timestamp

    def __str__(self):
        return '[ Timestamp: %d ]' % self.timestamp

    def __repr__(self):
        return 'TimestampPacket(%d)' % self.timestamp

    def __eq__(self, other):
        return self.timestamp == other.timestamp

    def __ne__(self, other):
        return not (self == other)

    def export(self):
        return {
                'type_str': 'timestamp',
                'type': self.packet_type,
                'timestamp': self.timestamp,
                'bits': self.bits.to01(),
                }

    def from_dict(self, d):
        ''' Inverse of export - modify packet based on dict '''
        if 'type' in d and d['type'] != self.packet_type:
            raise ValueError('invalid packet type for TimestampPacket')
        for key, value in d.items():
            if key == 'type':
                self.packet_type = value
            elif key == 'type_str':
                continue
            elif key == 'bits':
                self.bits = bitarray(value)
            else:
                setattr(self, key, value)

    @property
    def bits(self):
        return bah.fromuint(self.timestamp, self.size)

    @bits.setter
    def bits(self, value):
        self.timestamp = bah.touint(value)

    def bytes(self):
        return struct.pack('Q', self.timestamp)



'''
Quickstart commands for test boards
'''

from __future__ import absolute_import
from __future__ import print_function
import sys

from . import Controller
from .io import ZMQ_IO
from .logger.h5_logger import HDF5Logger

## For interactive mode
VERSION = sys.version_info
if VERSION[0] < 3:
    input = raw_input
##

def silence_chips(controller, interactive):
    '''Silence all chips in controller'''
    for chip_key in controller.chips:
        chip = controller.get_chip(chip_key)
        if interactive:
            print('Silencing chip %d' % chip.chip_id)
        chip.config.load('chip/quiet.json')
        controller.write_configuration(chip_key)
        if interactive:
            input('Just silenced chip %d. <enter> when ready.\n' % chip.chip_id)
    return

def set_config_physics(controller, interactive):
    '''Set the chips for the default physics configuration'''
    for chip_key in controller.chips:
        chip = controller.get_chip(chip_key)
        if interactive:
            x = input('Configuring chip %d. <enter> to continue, q to quit' % chip.chip_id)
            if x == 'q':
                break
        chip.config.load('chip/physics.json')
        controller.write_configuration(chip_key)
        print('configured {}'.format(str(chip)))
    return

def quickcontroller(board='pcb-5', interactive=False, io=None, logger=None,
    log_filepath=None):
    '''Quick jump through all controller creation and config steps'''
    controller_config = 'controller/{}_chip_info.json'.format(board)
    if io is None:
        io_config = 'io/daq-srv1.json'
        io = ZMQ_IO(io_config)
    if logger is None:
        logger = HDF5Logger(filename=log_filepath)
        logger.open()
    controller = Controller()
    controller.io = io
    controller.logger = logger
    controller.load(controller_config)
    silence_chips(controller, interactive)
    set_config_physics(controller, interactive)
    return controller

# Short-cut handle
qc = quickcontroller




from __future__ import print_function
import time

from larpix.logger import Logger

class StdoutLogger(Logger):
    '''
    The StdoutLogger is logger class that acts as a test logger class. All objects
    are displayed according to their string representation and routed to stdout.

    :param buffer_length: how many data messages to hang on to before flushing buffer to stdout
    :param mode: how logger file should be opened (not implemented in ``StdoutLogger``)
    '''
    def __init__(self, filename=None, buffer_length=0, mode='wa',
            enabled=False):
        super(StdoutLogger, self).__init__(enabled=enabled)
        self.filename = filename
        self._buffer = []
        self.buffer_length = buffer_length

    def record(self, data, direction=0):
        '''
        Send the specified data to stdout

        :param data: list of data to be written to log
        :param direction: 0 if packets were sent to ASICs, 1 if packets
            were received from ASICs. optional, default=0
        '''
        if not self.is_enabled():
            return
        if not isinstance(data,list):
            raise ValueError('data must be a list')

        self._buffer += ['Record: {}'.format(str(data_obj)) for data_obj in data]

        if len(self._buffer) > self.buffer_length:
            self.flush()

    def flush(self):
        '''
        Flushes any held data
        '''
        for msg in self._buffer:
            print(msg)
        self._buffer = []



from __future__ import print_function

import pytest
import h5py

from larpix.larpix import (Packet_v1, Packet_v2, PacketCollection, TimestampPacket,
                           MessagePacket, Key, SyncPacket, TriggerPacket)
from larpix.format.hdf5format import (to_file, from_file,
        dtype_property_index_lookup)

@pytest.fixture
def tmpfile(tmpdir):
    return str(tmpdir.join('test.h5'))

@pytest.fixture
def data_packet():
    p = Packet_v1()
    p.packet_type = Packet_v1.DATA_PACKET
    p.chipid = 123
    p.channel = 7
    p.timestamp = 123456
    p.dataword = 120
    p.fifo_half_flag = 1
    p.assign_parity()
    p.chip_key = Key('1-2-123')
    p.direction = 1
    return p

@pytest.fixture
def config_read_packet():
    p = Packet_v1()
    p.packet_type = Packet_v1.CONFIG_READ_PACKET
    p.chipid = 123
    p.register_address = 10
    p.register_data = 23
    p.assign_parity()
    p.chip_key = Key('1-2-123')
    p.direction = 1
    return p

@pytest.fixture
def data_packet_v2():
    p = Packet_v2()
    p.packet_type = Packet_v2.DATA_PACKET
    p.chip_id = 123
    p.channel_id = 7
    p.timestamp = 123456
    p.dataword = 120
    p.shared_fifo = 1
    p.assign_parity()
    p.chip_key = Key('1-2-123')
    p.direction = 1
    return p

@pytest.fixture
def fifo_diagnostics_packet_v2():
    p = Packet_v2()
    p.enable_fifo_diagnostics = True
    p.packet_type = Packet_v2.DATA_PACKET
    p.chip_id = 123
    p.channel_id = 7
    p.timestamp = 123456
    p.dataword = 120
    p.shared_fifo = 1
    p.shared_fifo_events = 5
    p.local_fifo_events = 1
    p.assign_parity()
    p.chip_key = Key('1-2-123')
    p.direction = 1
    return p

@pytest.fixture
def config_read_packet_v2():
    p = Packet_v2()
    p.packet_type = Packet_v2.CONFIG_READ_PACKET
    p.chip_id = 123
    p.register_address = 10
    p.register_data = 23
    p.assign_parity()
    p.chip_key = Key('1-2-123')
    p.direction = 1
    return p

@pytest.fixture
def timestamp_packet():
    p = TimestampPacket(timestamp=12345)
    return p

@pytest.fixture
def message_packet():
    p = MessagePacket('Hello, World!', 1234567)
    return p

@pytest.fixture
def sync_packet():
    p = SyncPacket(io_group=1, sync_type=b'A', timestamp=123456, clk_source=1)
    return p

@pytest.fixture
def trigger_packet():
    p = TriggerPacket(io_group=1, trigger_type=b'0', timestamp=123456)
    return p

def test_to_file_v0_0_empty(tmpfile):
    to_file(tmpfile, [], version='0.0')
    f = h5py.File(tmpfile, 'r')
    assert f['_header']
    assert f['_header'].attrs['version']
    assert f['_header'].attrs['created']
    assert f['_header'].attrs['modified']
    assert f['raw_packet']

def test_to_file_v0_0_data_packet(tmpfile, data_packet):
    to_file(tmpfile, [data_packet], version='0.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['raw_packet']) == 1
    row = f['raw_packet'][0]
    props = dtype_property_index_lookup['0.0']['raw_packet']
    new_packet = Packet_v1()
    new_packet.chip_key = row[props['chip_key']]
    new_packet.packet_type = row[props['type']]
    new_packet.chipid = row[props['chipid']]
    new_packet.parity_bit_value = row[props['parity']]
    new_packet.channel = row[props['channel']]
    new_packet.timestamp = row[props['timestamp']]
    new_packet.dataword = row[props['adc_counts']]
    new_packet.fifo_half_flag = row[props['fifo_half']]
    new_packet.fifo_full_flag = row[props['fifo_full']]
    assert new_packet == data_packet

def test_to_file_v0_0_config_read_packet(tmpfile, config_read_packet):
    to_file(tmpfile, [config_read_packet], version='0.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['raw_packet']) == 1
    row = f['raw_packet'][0]
    props = dtype_property_index_lookup['0.0']['raw_packet']
    new_packet = Packet_v1()
    new_packet.chip_key = row[props['chip_key']]
    new_packet.packet_type = row[props['type']]
    new_packet.chipid = row[props['chipid']]
    new_packet.parity_bit_value = row[props['parity']]
    new_packet.register_address = row[props['register']]
    new_packet.register_data = row[props['value']]
    assert new_packet == config_read_packet

def test_to_file_v0_0_timestamp_packet(tmpfile, timestamp_packet):
    to_file(tmpfile, [timestamp_packet], version='0.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['raw_packet']) == 1
    row = f['raw_packet'][0]
    props = dtype_property_index_lookup['0.0']['raw_packet']
    assert row[props['type']] == 4  # packet_type
    new_packet = TimestampPacket()
    new_packet.timestamp = row[props['timestamp']]
    assert new_packet == timestamp_packet

def test_to_file_v0_0_many_packets(tmpfile, data_packet, config_read_packet,
        timestamp_packet):
    to_file(tmpfile, [data_packet, config_read_packet,
        timestamp_packet], version='0.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['raw_packet']) == 3

def test_from_file_v0_0_many_packets(tmpfile, data_packet,
        config_read_packet, timestamp_packet):
    packets = [data_packet, config_read_packet, timestamp_packet]
    to_file(tmpfile, packets, version='0.0')
    new_packets_dict = from_file(tmpfile, version='0.0')
    assert new_packets_dict['created']
    assert new_packets_dict['version']
    assert new_packets_dict['modified']
    new_packets = new_packets_dict['packets']
    assert new_packets[0] == data_packet
    assert new_packets[1] == config_read_packet
    assert new_packets[2] == timestamp_packet

def test_to_file_v1_0_empty(tmpfile):
    to_file(tmpfile, [], version='1.0')
    f = h5py.File(tmpfile, 'r')
    assert f['_header']
    assert f['_header'].attrs['version']
    assert f['_header'].attrs['created']
    assert f['_header'].attrs['modified']
    assert f['packets']

def test_to_file_v1_0_data_packet(tmpfile, data_packet):
    to_file(tmpfile, [data_packet], version='1.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 1
    row = f['packets'][0]
    props = dtype_property_index_lookup['1.0']['packets']
    new_packet = Packet_v1()
    new_packet.chip_key = row[props['chip_key']]
    new_packet.packet_type = row[props['type']]
    new_packet.chipid = row[props['chipid']]
    new_packet.parity_bit_value = row[props['parity']]
    new_packet.channel = row[props['channel']]
    new_packet.timestamp = row[props['timestamp']]
    new_packet.dataword = row[props['adc_counts']]
    new_packet.fifo_half_flag = row[props['fifo_half']]
    new_packet.fifo_full_flag = row[props['fifo_full']]
    new_packet.direction = row[props['direction']]
    assert new_packet == data_packet

def test_to_file_v1_0_config_read_packet(tmpfile, config_read_packet):
    to_file(tmpfile, [config_read_packet], version='1.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 1
    row = f['packets'][0]
    props = dtype_property_index_lookup['1.0']['packets']
    new_packet = Packet_v1()
    new_packet.chip_key = row[props['chip_key']]
    new_packet.packet_type = row[props['type']]
    new_packet.chipid = row[props['chipid']]
    new_packet.parity_bit_value = row[props['parity']]
    new_packet.register_address = row[props['register']]
    new_packet.register_data = row[props['value']]
    new_packet.direction = row[props['direction']]
    assert new_packet == config_read_packet

def test_to_file_v1_0_timestamp_packet(tmpfile, timestamp_packet):
    to_file(tmpfile, [timestamp_packet], version='1.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 1
    row = f['packets'][0]
    props = dtype_property_index_lookup['1.0']['packets']
    assert row[1] == 4  # packet_type
    new_packet = TimestampPacket()
    new_packet.timestamp = row[props['timestamp']]
    assert new_packet == timestamp_packet

def test_to_file_v1_0_message_packet(tmpfile, message_packet,
        data_packet):
    to_file(tmpfile, [data_packet, message_packet], version='1.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 2
    row = f['packets'][1]
    props = dtype_property_index_lookup['1.0']['packets']
    message_props = dtype_property_index_lookup['1.0']['messages']
    assert row[props['type']] == 5
    assert row[props['counter']] == 0
    assert row[props['timestamp']] == 1234567
    assert len(f['messages']) == 1
    message_row = f['messages'][0]
    assert message_row[message_props['index']] == 0
    assert message_row[message_props['message']] == b'Hello, World!'
    assert message_row[message_props['timestamp']] == 1234567

def test_to_file_v1_0_many_packets(tmpfile, data_packet, config_read_packet,
        timestamp_packet, message_packet):
    to_file(tmpfile, [data_packet, config_read_packet,
        timestamp_packet, message_packet], version='1.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 4
    assert len(f['messages']) == 1

def test_from_file_v1_0_many_packets(tmpfile, data_packet,
        config_read_packet, timestamp_packet, message_packet):
    packets = [data_packet, config_read_packet, timestamp_packet,
            message_packet]
    to_file(tmpfile, packets, version='1.0')
    new_packets_dict = from_file(tmpfile)
    assert new_packets_dict['created']
    assert new_packets_dict['version']
    assert new_packets_dict['modified']
    new_packets = new_packets_dict['packets']
    assert new_packets[0] == data_packet
    assert new_packets[1] == config_read_packet
    assert new_packets[2] == timestamp_packet
    assert new_packets[3] == message_packet

def test_to_file_v2_0_empty(tmpfile):
    to_file(tmpfile, [], version='2.0')
    f = h5py.File(tmpfile, 'r')
    assert f['_header']
    assert f['_header'].attrs['version']
    assert f['_header'].attrs['created']
    assert f['_header'].attrs['modified']
    assert f['packets']

def test_to_file_v2_0_data_packet(tmpfile, data_packet_v2):
    to_file(tmpfile, [data_packet_v2], version='2.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 1
    row = f['packets'][0]
    props = dtype_property_index_lookup['2.0']['packets']
    new_packet = Packet_v2()
    new_packet.chip_key = Key(row[props['io_group']], row[props['io_channel']], row[props['chip_id']])
    new_packet.packet_type = row[props['packet_type']]
    new_packet.trigger_type = row[props['trigger_type']]
    new_packet.chip_id = row[props['chip_id']]
    new_packet.parity = row[props['parity']]
    new_packet.channel_id = row[props['channel_id']]
    new_packet.timestamp = row[props['timestamp']]
    new_packet.dataword = row[props['dataword']]
    new_packet.local_fifo = row[props['local_fifo']]
    new_packet.shared_fifo = row[props['shared_fifo']]
    new_packet.direction = row[props['direction']]
    assert new_packet.timestamp == data_packet_v2.timestamp
    assert new_packet == data_packet_v2

def test_to_file_v2_0_config_read_packet(tmpfile, config_read_packet_v2):
    to_file(tmpfile, [config_read_packet_v2], version='2.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 1
    row = f['packets'][0]
    props = dtype_property_index_lookup['2.0']['packets']
    new_packet = Packet_v2()
    new_packet.chip_key = Key(row[props['io_group']], row[props['io_channel']], row[props['chip_id']])
    new_packet.packet_type = row[props['packet_type']]
    new_packet.parity = row[props['parity']]
    new_packet.register_address = row[props['register_address']]
    new_packet.register_data = row[props['register_data']]
    new_packet.direction = row[props['direction']]
    print(new_packet)
    print(config_read_packet_v2)
    assert new_packet == config_read_packet_v2

def test_to_file_v2_0_timestamp_packet(tmpfile, timestamp_packet):
    to_file(tmpfile, [timestamp_packet], version='2.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 1
    row = f['packets'][0]
    props = dtype_property_index_lookup['2.0']['packets']
    assert row[3] == 4  # packet_type
    new_packet = TimestampPacket()
    new_packet.timestamp = row[props['timestamp']]
    assert new_packet == timestamp_packet

def test_to_file_v2_0_message_packet(tmpfile, message_packet,
        data_packet_v2):
    to_file(tmpfile, [data_packet_v2, message_packet], version='2.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 2
    row = f['packets'][1]
    props = dtype_property_index_lookup['2.0']['packets']
    message_props = dtype_property_index_lookup['2.0']['messages']
    assert row[props['packet_type']] == 5
    assert row[props['timestamp']] == 1234567
    assert len(f['messages']) == 1
    message_row = f['messages'][0]
    assert message_row[message_props['index']] == 0
    assert message_row[message_props['message']] == b'Hello, World!'
    assert message_row[message_props['timestamp']] == 1234567

def test_to_file_v2_0_many_packets(tmpfile, data_packet_v2, config_read_packet_v2,
        timestamp_packet, message_packet):
    to_file(tmpfile, [data_packet_v2, config_read_packet_v2,
        timestamp_packet, message_packet], version='2.0')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 4
    assert len(f['messages']) == 1

def test_from_file_v2_0_many_packets(tmpfile, data_packet_v2,
        config_read_packet_v2, timestamp_packet, message_packet):
    packets = [data_packet_v2, config_read_packet_v2, timestamp_packet,
            message_packet]
    to_file(tmpfile, packets, version='2.0')
    new_packets_dict = from_file(tmpfile)
    assert new_packets_dict['created']
    assert new_packets_dict['version']
    assert new_packets_dict['modified']
    new_packets = new_packets_dict['packets']
    _ = [print(str(packet) +'\n' + str(packets[i])) for i,packet in enumerate(new_packets)]
    assert new_packets[0] == data_packet_v2
    assert new_packets[1] == config_read_packet_v2
    assert new_packets[2] == timestamp_packet
    assert new_packets[3] == message_packet

def test_to_file_v2_2_many_packets(tmpfile, data_packet_v2, config_read_packet_v2,
                                   timestamp_packet, message_packet, sync_packet,
                                   trigger_packet):
    to_file(tmpfile, [data_packet_v2, config_read_packet_v2,
                      timestamp_packet, message_packet, sync_packet,
                      trigger_packet],
            version='2.2')
    f = h5py.File(tmpfile, 'r')
    assert len(f['packets']) == 6
    assert len(f['messages']) == 1

def test_from_file_v2_2_many_packets(tmpfile, data_packet_v2,
                                     config_read_packet_v2, timestamp_packet,
                                     message_packet, sync_packet, trigger_packet):
    packets = [data_packet_v2, config_read_packet_v2, timestamp_packet,
               message_packet, sync_packet, trigger_packet]
    to_file(tmpfile, packets, version='2.2')
    new_packets_dict = from_file(tmpfile)
    assert new_packets_dict['created']
    assert new_packets_dict['version']
    assert new_packets_dict['modified']
    new_packets = new_packets_dict['packets']
    _ = [print(str(packet) +'\n' + str(packets[i])) for i,packet in enumerate(new_packets)]
    assert new_packets[0] == data_packet_v2
    assert new_packets[1] == config_read_packet_v2
    assert new_packets[2] == timestamp_packet
    assert new_packets[3] == message_packet
    assert new_packets[4] == sync_packet
    assert new_packets[5] == trigger_packet

def test_from_file_incompatible(tmpfile):
    to_file(tmpfile, [], version='0.0')
    with pytest.raises(RuntimeError):
        from_file(tmpfile, version='1.0')
        pytest.fail('Should identify incompatible version')



'''
This module contains controller configuration files, used to specify the physical
asics present and their network structure

V2 Configuration
----------------

The v2 configuration file is a standard JSON file structured as follows:

.. parsed-literal::
    {
        "_config_type": "controller",
        "_include": [<optional list of files to inherit from>],
        "name": <string identifier for configuration ID, e.g. "pcb-3", or "acube-2x2">,
        "layout": <string identifier for layout version, e.g. "1.2.0">,
        "asic_version": 2,
        "network": {
            "miso_us_uart_map": [<uart channel for position in array, required>, ...],
            "miso_ds_uart_map": [<uart channel for position in array, required>, ...],
            "mosi_uart_map": [<uart channel for position in array, required>, ...],
            "<io_group>": {
                "miso_us_uart_map": [<uart channels for io_group, optional>, ...],
                ...
                "<io_channel>": {
                    "miso_us_uart_map": [<uart channels for io_group, optional>, ...],
                    ...
                    "nodes": [
                        {
                            "chip_id": <chip id of first chip to load into network>,
                            "miso_us": [<chip_ids of upstream chips, null if no upstream chip, optional>],
                            "miso_ds": [<chip_ids of downstream chips, optional>],
                            "mosi": [<chip_ids of chips to listen to, optional>],
                            "root": <bool indicating if this is a root node, optional>,
                            "miso_us_uart_map": [<uart channels for chip, optional>, ...],
                            ...
                        },
                        ...
                    ]
                },
                ...
            },
            ...
        }
    }

The structure allows for a lot of flexibility in the network configuration, while
maintaining a relatively terse format for simple configurations. Generally, a
Hydra network can be described using three directed graphs which we call
miso upstream (miso_us), miso downstream (miso_ds), and mosi. The miso_us graph
represents the flow of outgoing upstream packets (i.e. the direction of configuration
write packets). The miso_ds graph represents the flow of outgoing downstream packet
(i.e. the direction of configuration read and data packets). The mosi graph
represents the flow of incoming packets of either type. The overall LArPix
network is then defined with a collection of Hydra networks, each associated
with a specific io_group and io_channel. The io_group, io_channel pair must
be unique for each Hydra network.

For the configuration file, the basic logic for each hydra network is that for each
chip, you must specify a chip id. Links within the directed graphs are then specified via
4-item arrays. A simplified example is provided below:

.. parsed-literal::
    "network": {
        "1": {
            "2": {
                "chips": [
                    {
                        "chip_id": "ext",
                        "miso_us": [null,null,2,null]
                    }
                    {
                        "chip_id": 2,
                        "miso_us": [3,null,null,null],
                        "root": true
                    },
                    {
                        "chip_id": 3
                    }
                ]
            }
        }
        "miso_us_uart_map": [0,1,2,3],
        "miso_ds_uart_map": [2,3,0,1],
        "mosi_uart_map": [0,1,2,3]
    }

In this example, we define a single 2-chip Hydra network with io_group=1
and io_channel=2. The "chips" array is filled with dicts representing each
chip in the network. Within the chip specification, the "chip_id" field sets
the chip_id configuration register for that chip and must be unique within
each io_group, io_channel sub-network. The "miso_us" array lists the chip_ids
that should receive upstream packets from this chip's miso_us channels. Using
the miso_us_uart map specified in the example, index=0 refers to a connection on
miso uart channel 0, index=1 refers to uart channel 1, etc. The miso_us should
be specified for each chip that is linked to other chips.

'Dummy' nodes can be specified by using a string for the "chip_id", rather than
an integer. These represent network links to non-chip object, e.g. an fpga, etc.
In general, you must specify one dummy link that points to at least one chip on
the "miso_us" network.

In each network there must be at least 1 "root" node. The node must have the
"root" field set to true. This determines the order in which chips will be configured.
The uart channel of each position in miso_ds network is inferred by the
miso_ds_uart_map, so for the example configuration uart0 of chip 2 will send
downstream packets. All other network links will be inferred from these fields.

If you'd like to manually specify miso_us links for a chip,
you can specify their chip ids in a list of 4 upstream channels. Values of null
are ignored, and non-integer values are interpreted as placeholders (i.e. no
physical chip, but we'd like to enable the miso_us channel for some reason or
another). The physical uart to enable/disable
during the configuration of the network is gleaned from the "miso_us_uart_map".
The "miso_us_uart_map" can be specified at any layer of the configuration (io
group, io channel, or chip) and will override the values specified at the
network level.

A similar logic applies for the miso_ds, however by default, this network
uses the converse of the miso_us configuration (with the uart channel specified
by "miso_ds_uart_map"). For example, to declare a miso_us/miso_ds connection
between two chips, you may simple declare:

    "chips": [
        {
            "chip_id": 2,
            "miso_us":[3,null,null,null],
            "miso_us_uart_map": [0,1,2,3]
        },
        {
            "chip_id": 3,
            "miso_ds_uart_map": [2,3,0,1]
        }
    ]

This will initialize chip 2 with the miso_us enabled on uart channel 0 and
chip 3 with the miso_ds enabled on uart channel 2. Placeholders (such as 'ext')
be used with miso_ds as well.

For mosi, the default behavior is to use the converse of both the
miso_us and miso_ds graphs using the same uarts. E.g. the above example would
have chip 2 with mosi enabled on uart channel 0 and chip 3 with mosi enabled on
uart channel 2. You can override this behavior by setting the "mosi" field to
an array of the chips to link on the mosi graph. To override which physical uart
to used, the "mosi_uart_map" is used in a similar fashion as miso_us.

To properly create the uart maps, it requires knowing the routing of mosi/miso
channels between the chips. The miso_us_uart_map determines which uart miso
channel to enable if a miso upstream graph edge originates from the given chip.
The miso_ds_uart_map determines which uart miso channel to enable if an a miso
downstream graph edge originates from the given chip. And the mosi_uart_map
determines which uart channel to enable if a mosi graph edge ends on the given
chip. E.g.::

    "miso_us_uart_map": [3,0,1,2],
    "miso_ds_uart_map": [1,2,3,0],
    "mosi_uart_map": [0,3,2,1]

declares that a link created at index 1 will use uart channel 0 (if it is a miso_us link),
channel 2 (if it is a miso_ds link), or channel 3 (if it is a mosi link).

Finally, the ``'root'`` field is used to specify special 'root' nodes within the
network. These are the nodes that will be configured first when initializing a
network (with the next nodes order of configuration determined via the mosi_us
configuration). If not specified, the node is assumed to not be a root node.

The top-level "_include" field can be used to specify files to inherit from.
All fields of the configuration down to the "chips" array level can be inherited.
A standard use for this would be to specify each hydra network for a given channel
independently, and inherit from all of the files in a single configuration
file. E.g.::

    {
        "_config_type": "controller",
        "_include": [
            "network-1-1.json",
            "network-1-2.json",
            "network-1-3.json",
            "network-1-4.json",
            "network-2-1.json",
            "network-2-2.json",
            "network-2-3.json",
            "network-2-4.json"
            ],
        "name": "multi-io-group-network"
        "layout": null
    }


V1 Configuration
----------------

The v1 configuration file is a standard JSON file structured as follows:

.. parsed-literal::
    {
        "name": <string identifier for PCB ID, e.g. "pcb-3">,
        "layout": <string identifier for layout version, e.g. "1.2.0">,
        "asic_version": 1,
        "chip_list": [<A list of chip keys, one for each chip, in daisy-chain order>]
    }

All fields are necessary.

See the larpix-geometry documentation
<https://github.com/samkohn/larpix-geometry> for a listing of all the
layout versions.
'''
pass



from bitarray import bitarray
import os
import errno
import functools
from collections import OrderedDict

from .. import bitarrayhelper as bah
from .. import configs
from . import BaseConfiguration, _Smart_List

class Configuration_v2(BaseConfiguration):
    '''
    Represents the desired configuration state of a LArPix v2 chip.

    Each register name is available as its own attribute for inspecting and
    setting the value of the corresponding register.

    Certain configuration values are set channel-by-channel. These are
    represented by a list of values. For example:

        >>> conf.pixel_trim_dac[2:5]
        [16, 16, 16]
        >>> conf.channel_mask[20] = 1
        >>> conf.external_trigger_mask = [0] * 64

    Additionally, other configuration values take up more than or less
    than one complete register. These are still set by referencing the
    appropriate name. For example, ``cross_trigger_mode`` shares a
    register with a few other values, and adjusting the value of the
    ``cross_trigger_mode`` attribute will leave the other values
    unchanged.

    Each register name can cover more than one 'physical' register depending on
    the size of the data it holds. You can see which physical registers a
    given register name corresponds to by using the `register_map` attribute, e.g.::

        >>> conf.register_map['digital_threshold']  # 64 registers, 1 per channel
        range(173, 237)
        >>> conf.register_map['enable_dynamic_reset']  # Register 170
        range(170, 171)
        >>> conf.register_map['enable_min_delta_adc']  # Shares register 170
        range(170, 171)

    '''

    asic_version = 2
    default_configuration_file = 'chip/default_v2.json'
    num_registers = 237
    num_bits = 1896
    num_channels = 64
    # Additional class properties regarding configuration registers are set at the end of the file.

    _endian = 'little'

    def __init__(self):
        # Note: properties, getters and setters are constructed after this class definition at the bottom of the file.
        super(Configuration_v2, self).__init__()
        return

    def all_data(self, endian='little'):
        bits = []
        for register_name in self.register_names:
            register_data =  getattr(self, register_name+'_data')
            for register_addr, register_bits in register_data:
                if len(bits) == register_addr:
                    if endian[0] == 'l':
                        bits += [register_bits]
                    else:
                        bits += [register_bits[::-1]]
        return bits

    def from_dict_registers(self, d, endian='little'):
        '''
        Load in the configuration specified by a dict of (register,
        value) pairs.

        '''
        for address, value in d.items():
            register_names = self.register_map_inv[address]
            for register_name in register_names:
                setattr(self, register_names[0] + '_data', (address, bah.fromuint(value,8,endian=endian)))
        return

    def _is_register_value_pair(self, item):
        '''
        Helper function to determine if item is a register, bitarray pair

        '''
        if isinstance(item, tuple) and len(item) == 2 and \
                isinstance(item[0], int) and isinstance(item[1], bitarray)\
                and len(item[1]) == 8 and item[0] >= 0 and item[0] < self.num_registers:
            return True
        return False

### Stuff for generating the v2 configuration properties

## Getter function formulas
#
def _basic_getter(register_name):
    '''
    Function formula for getting a named register

    '''
    def basic_getter_func(self):
        return getattr(self, '_'+register_name)
    return basic_getter_func
# /Getter function formulas

## Setter function formulas
#
def _basic_setter(register_name):
    '''
    Function formula for setting a named register

    '''
    def basic_setter_func(self, value):
        setattr(self, '_'+register_name, value)
    return basic_setter_func

def _list_setter(register_name, min_value, max_value):
    '''
    Function formula for setting a named register

    '''
    def list_setter_func(self, value):
        setattr(self, '_'+register_name, _Smart_List(value, min_value, max_value))
    return list_setter_func

# /Setter function formulas

## Data getter function formulas
#
def _basic_data_getter(register_name):
    '''
    Function formula for getting a "simple" register's data
    A simple register has a one-to-many relationship between the register
    name and register addresses.

    Returns a list of register_addr, bitarray pairs

    '''
    def basic_data_getter_func(self):
        register_range = self.register_map[register_name]
        value = getattr(self, register_name)
        all_data = bah.fromuint(value, 8*(register_range[-1]-register_range[0]+1), endian=self._endian)
        return_data = [(register, all_data[idx*8:idx*8+8]) for idx, register in enumerate(register_range)]
        return return_data
    return basic_data_getter_func

def _list_data_getter(register_name, n_bits):
    '''
    Function formula for getting a list-like register's data
    A list-like register contains repeated functionality every ``n_bits``
    over 1 or more register addresses.

    Returns a list of register_addr, bitarray pairs

    '''
    def list_data_getter_func(self):
        register_range = self.register_map[register_name]
        values = getattr(self, register_name)
        all_data = bitarray([])
        for value in values:
            all_data += bah.fromuint(value, n_bits, endian=self._endian)
        if len(all_data) < 8:
            all_data += bitarray([0]*(8-len(all_data)%8)) # pad up to full register
        return [(register_addr, all_data[idx*8:idx*8+8]) for idx, register_addr in enumerate(register_range)]
    return list_data_getter_func

def _compound_data_getter(registers):
    '''
    Function formula for getting a compound register's data
    A compound register has a one-to-many mapping of register address to
    register names.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)

    Returns a list of register, bitarray pairs

    '''
    def compound_data_getter_func(self):
        bits = bitarray([0]*8)
        for register_name in registers:
            start_bit, end_bit = self.bit_map[register_name]
            bits[start_bit%8:end_bit-start_bit+start_bit%8] = bah.fromuint(getattr(self, register_name), end_bit-start_bit, endian=self._endian)
        return [(self.register_map[registers[0]][0], bits)]
    return compound_data_getter_func

def _compound_list_data_getter(registers, n_bits):
    '''
    Function formula for getting a compound list register's data
    A compound list register has a one-to-many mapping of register
    addresses to register names, but with repeating functionality within
    each register name of size ``n_bits``. For example, you might store two
    1bit arrays of length 2 within the same register address.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)

    Returns a list of register, bitarray pairs

    '''
    def compound_list_data_getter_func(self):
        bits = bitarray([])
        for register_name in registers:
            values = getattr(self, register_name)
            for value in values:
                bits += bah.fromuint(value, n_bits, endian=self._endian)
        if len(bits)%8 != 0:
            bits += bitarray([0]*(8-len(bits)%8)) # pad up to full register
        return [(self.register_map[registers[0]][0], bits)]
    return compound_list_data_getter_func
# /Data getter function formulas

## Data setter function formulas
#
def _basic_data_setter(register_name):
    '''
    Function formula for setting a simple register's data
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def basic_data_setter_func(self, value):
        if self._is_register_value_pair(value):
            # regenerate complete list of bits and set
            set_register, set_bits = value
            data = getattr(self, register_name+'_data')
            all_bits = bitarray([])
            for register, bits in data:
                if register == set_register:
                    all_bits += set_bits
                else:
                    all_bits += bits
            start_bit, end_bit = self.bit_map[register_name]
            setattr(self, register_name+'_data', all_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            # use all bits to set values
            bits = value
            value = bah.touint(bits, endian=self._endian)
            setattr(self, register_name, value)
    return basic_data_setter_func

def _list_data_setter(register_name, n_bits, min_value, max_value):
    '''
    Function formula for setting a list-like register's data
    A list-like register contains repeated functionality every ``n_bits``
    over 1 or more register addresses.
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def list_data_setter_func(self, values):
        if self._is_register_value_pair(values):
            # regenerate complete list of bits and set
            set_register, set_bits = values
            data = getattr(self, register_name+'_data')
            all_bits = bitarray([])
            for register, bits in data:
                if register == set_register:
                    all_bits += set_bits
                else:
                    all_bits += bits
            start_bit, end_bit = self.bit_map[register_name]
            setattr(self, register_name+'_data', all_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            # use all bits to set values
            bits = values
            item_values = [bah.touint(bits[idx:idx+n_bits], endian=self._endian) for idx in range(0, len(bits), n_bits)]
            setattr(self, register_name, _Smart_List(item_values, min_value, max_value))
    return list_data_setter_func

def _compound_data_setter(registers, register_name):
    '''
    Function formula for setting a compound register's data
    A compound register has a one-to-many mapping of register address to
    register names.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def compound_data_setter_func(self, value):
        if self._is_register_value_pair(value):
            set_register_addr, set_bits = value
            for register in registers:
                start_bit, end_bit = self.bit_map[register]
                setattr(self, register + '_data', set_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            # use all bits to set values
            set_bits = value
            setattr(self, register_name, bah.touint(set_bits, endian=self._endian))
    return compound_data_setter_func

def _compound_list_data_setter(registers, register_name, n_bits, min_value, max_value):
    '''
    Function formula for setting a compound list register's data
    A compound list register has a one-to-many mapping of register address to
    register names with repeating functionality every ``n_bits``.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def compound_list_data_setter_func(self, value):
        if self._is_register_value_pair(value):
            set_register_addr, set_bits = value
            for register in registers:
                start_bit, end_bit = self.bit_map[register]
                setattr(self, register + '_data', set_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            set_bits = value
            values = [bah.touint(set_bits[idx:idx+n_bits], endian=self._endian) for idx in range(0,len(set_bits),n_bits)]
            setattr(self, register_name, _Smart_List(values, min_value, max_value))
    return compound_list_data_setter_func
# /Data setter function formulas

## Value validation function formulas
#
def _value_validator(value_types, min_value, max_value):
    '''
    Function formula for validating a register that contains a single value
    Accepts values of types ``value_types``, greater than  ``min_value``,
    and less than ``max_value``.

    '''
    def value_validator(func):
        @functools.wraps(func)
        def value_validated_func(self, value):
            if not isinstance(value, value_types):
                raise TypeError('value must be of type {}'.format(value_types))
            if value > max_value or value < min_value:
                raise ValueError('value must be between {} and {}'.format(min_value,max_value))
            return func(self, value)
        return value_validated_func
    return value_validator

def _list_validator(value_types, min_value, max_value, n_values):
    '''
    Function formula for validating a register that contains list-like data
    Accepts lists of length ``n_values`` with objects with types
    ``value_types``, greater than  ``min_value``, and less than ``max_value``.

    '''
    def list_validator(func):
        @functools.wraps(func)
        def list_validated_func(self, values):
            if not isinstance(values, (list, _Smart_List, tuple)):
                raise TypeError('argument must be a list or tuple')
            if len(values) != n_values:
                raise ValueError('length of list must be {}'.format(n_values))
            if any([not isinstance(value, value_types) for value in values]):
                raise TypeError('values must be of type {}'.format(value_types))
            if any([value > max_value or value < min_value for value in values]):
                raise ValueError('values must be between {} and {}'.format(min_value,max_value))
            return func(self, values)
        return list_validated_func
    return list_validator

def _data_validator(register_name):
    '''
    Function formula for validating setting a register via a register
    address, bitarray pair or direct bitarray

    '''
    def data_validator(func):
        @functools.wraps(func)
        def data_validated_func(self, value):
            print('validate',register_name,value)
            if self._is_register_value_pair(value):
                register_addr, bits  = value
                if register_addr < self.register_map[register_name][0] \
                or register_addr > self.register_map[register_name][-1] \
                or len(bits) != 8:
                    raise ValueError('invalid register, value pair {}'
                        ' for register {}'.format(value,register_name))
                return func(self, value)
            elif isinstance(value, bitarray):
                bits = value
                start_bit, stop_bit = self.bit_map[register_name]
                if len(bits) != stop_bit - start_bit:
                    raise ValueError('invalid bitarray {}'
                        ' for register {}'.format(value,register_name))
                return func(self, value)
            else:
                raise TypeError('{} data must be assigned with '
                    'register, value pair or bitarray'.format(register_name))
        return data_validated_func
    return data_validator
# /Value validation function formulas

## Property function formulas
#
def _basic_property(name, registers, bits, types, min, max):
    docstring = '''
        simple value property

        registers: ``{}``

        valid types: ``{}``

        value range: ``{}`` to ``{}``
        '''.format(registers, types, min, max)
    return (
        property(
            _basic_getter(name),
            _value_validator(types,min,max)(
                _basic_setter(name)),
            doc=docstring),
        property(
            _basic_data_getter(name),
            _data_validator(name)(
                _basic_data_setter(name))))

def _list_property(name, registers, bits, types, min, max, length, n_bits):
    docstring = '''
        list-like property

        list length: ``{}``

        registers: ``{}``

        element valid types: ``{}``

        element value range: ``{}`` to ``{}``
        '''.format(length, registers, types, min, max)
    return (
        property(
            _basic_getter(name),
            _list_validator(types,min,max,length)(
                _list_setter(name,min,max)),
            doc=docstring),
        property(
            _list_data_getter(name,n_bits),
            _data_validator(name)(
                _list_data_setter(name,n_bits,min,max))))

def _compound_property(name, registers, bits, names, types, min, max):
    docstring = '''
        compound property

        registers: ``{}``

        bits: ``{}``

        valid types: ``{}``

        value range: ``{}`` to ``{}``

        shares a register with: ``{}``
        '''.format(registers, [bit%8 for bit in bits], types, min, max, [n for n in names if not n == name])
    return (
        property(
            _basic_getter(name),
            _value_validator(types,min,max)(
                _basic_setter(name)),
            doc=docstring
            ),
        property(
            _compound_data_getter(names),
            _data_validator(name)(
                _compound_data_setter(names, name))))

def _compound_list_property(name, registers, bits, names, types, min, max, length, n_bits):
    docstring = '''
        compound list-like property

        list length: ``{}``

        registers: ``{}``

        element valid types: ``{}``

        element value range: ``{}`` to ``{}``

        shares a register with: ``{}``
        '''.format(length, registers, types, min, max, [n for n in names if not n == name])
    return (
        property(
            _basic_getter(name),
            _list_validator(types,min,max,length)(
                _list_setter(name,min,max)),
            doc=docstring
            ),
        property(
            _compound_list_data_getter(names,n_bits),
            _data_validator(name)(
                _compound_list_data_setter(names,name,n_bits,min,max))))
# /Property function formulas

## Set up property info
#
_property_configuration = OrderedDict([
        ('pixel_trim_dac',
            (_list_property, (int, 0, 31, Configuration_v2.num_channels, 8), (0,512))),
        ('threshold_global',
            (_basic_property, (int, 0, 255), (512, 520))),
        ('csa_gain',
            (_compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (520,521))),
        ('csa_bypass_enable',
            (_compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (521,522))),
        ('bypass_caps_en',
            (_compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (522,523))),
        ('csa_enable',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (528, 592))),
        ('ibias_tdac',
            (_basic_property, (int, 0, 15), (592, 596))),
        ('ibias_comp',
            (_basic_property, (int, 0, 15), (600, 604))),
        ('ibias_buffer',
            (_basic_property, (int, 0, 15), (608, 612))),
        ('ibias_csa',
            (_basic_property, (int, 0, 15), (616, 620))),
        ('ibias_vref_buffer',
            (_basic_property, (int, 0, 15), (624, 628))),
        ('ibias_vcm_buffer',
            (_basic_property, (int, 0, 15), (632, 636))),
        ('ibias_tpulse',
            (_basic_property, (int, 0, 15), (640, 644))),
        ('ref_current_trim',
            (_compound_property, (['ref_current_trim','override_ref','ref_kickstart'], int, 0, 31), (648, 653))),
        ('override_ref',
            (_compound_property, (['ref_current_trim','override_ref','ref_kickstart'], (int,bool), 0, 1), (653, 654))),
        ('ref_kickstart',
            (_compound_property, (['ref_current_trim','override_ref','ref_kickstart'], (int,bool), 0, 1), (654, 655))),
        ('vref_dac',
            (_basic_property, (int, 0, 255), (656, 664))),
        ('vcm_dac',
            (_basic_property, (int, 0, 255), (664,672))),
        ('csa_bypass_select',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (672,736))),
        ('csa_monitor_select',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (736,800))),
        ('csa_testpulse_enable',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (800,864))),
        ('csa_testpulse_dac',
            (_basic_property, (int, 0, 255), (864,872))),
        ('current_monitor_bank0',
            (_list_property, ((int,bool), 0, 1, 4, 1), (872,876))),
        ('current_monitor_bank1',
            (_list_property, ((int,bool), 0, 1, 4, 1), (880,884))),
        ('current_monitor_bank2',
            (_list_property, ((int,bool), 0, 1, 4, 1), (888,892))),
        ('current_monitor_bank3',
            (_list_property, ((int,bool), 0, 1, 4, 1), (896,900))),
        ('voltage_monitor_bank0',
            (_list_property, ((int,bool), 0, 1, 3, 1), (904,907))),
        ('voltage_monitor_bank1',
            (_list_property, ((int,bool), 0, 1, 3, 1), (912,915))),
        ('voltage_monitor_bank2',
            (_list_property, ((int,bool), 0, 1, 3, 1), (920,923))),
        ('voltage_monitor_bank3',
            (_list_property, ((int,bool), 0, 1, 3, 1), (928,931))),
        ('voltage_monitor_refgen',
            (_list_property, ((int,bool), 0, 1, 8, 1), (936,944))),
        ('digital_monitor_enable',
            (_compound_property, (['digital_monitor_enable','digital_monitor_select'], (int,bool), 0, 1), (944,945))),
        ('digital_monitor_select',
            (_compound_property, (['digital_monitor_enable','digital_monitor_select'], (int,bool), 0, 10), (945,949))),
        ('digital_monitor_chan',
            (_basic_property, (int, 0, 63), (952,958))),
        ('slope_control0',
            (_compound_property, (['slope_control0', 'slope_control1'], int, 0, 15), (960,964))),
        ('slope_control1',
            (_compound_property, (['slope_control0', 'slope_control1'], int, 0, 15), (964,968))),
        ('slope_control2',
            (_compound_property, (['slope_control2', 'slope_control3'], int, 0, 15), (968,972))),
        ('slope_control3',
            (_compound_property, (['slope_control2', 'slope_control3'], int, 0, 15), (972,976))),
        ('chip_id',
            (_basic_property, (int, 0, 255), (976,984))),
        ('load_config_defaults',
            (_compound_property, (['load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl'], (int,bool), 0, 1), (985,986))),
        ('enable_fifo_diagnostics',
            (_compound_property, (['load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl'], (int,bool), 0, 1), (986,987))),
        ('clk_ctrl',
            (_compound_property, (['load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl'], (int), 0, 2), (987,989))),
        ('enable_miso_upstream',
            (_list_property, ((int,bool), 0, 1, 4, 1), (992,996))),
        ('enable_miso_downstream',
            (_compound_list_property, (['enable_miso_downstream', 'enable_miso_differential'], (int,bool), 0, 1, 4, 1), (1000,1004))),
        ('enable_miso_differential',
            (_compound_list_property, (['enable_miso_downstream', 'enable_miso_differential'], (int,bool), 0, 1, 4, 1), (1004,1008))),
        ('enable_mosi',
            (_list_property, ((int,bool), 0, 1, 4, 1), (1008,1012))),
        ('test_mode_uart0',
            (_compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1016,1018))),
        ('test_mode_uart1',
            (_compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1018,1020))),
        ('test_mode_uart2',
            (_compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1020,1022))),
        ('test_mode_uart3',
            (_compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1022,1024))),
        ('enable_cross_trigger',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1024,1025))),
        ('enable_periodic_reset',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1025,1026))),
        ('enable_rolling_periodic_reset',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1026,1027))),
        ('enable_periodic_trigger',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1027,1028))),
        ('enable_rolling_periodic_trigger',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1028,1029))),
        ('enable_periodic_trigger_veto',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1029,1030))),
        ('enable_hit_veto',
            (_compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1030,1031))),
        ('adc_hold_delay',
            (_basic_property, (int, 0, 15), (1032,1036))),
        ('adc_burst_length',
            (_basic_property, (int, 0, 255), (1040,1048))),
        ('channel_mask',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1048,1112))),
        ('external_trigger_mask',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1112,1176))),
        ('cross_trigger_mask',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1176,1240))),
        ('periodic_trigger_mask',
            (_list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1240,1304))),
        ('periodic_reset_cycles',
            (_basic_property, (int, 0, 2**24-1), (1304,1328))),
        ('periodic_trigger_cycles',
            (_basic_property, (int, 0, 2**32-1), (1328,1360))),
        ('enable_dynamic_reset',
            (_compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1360,1361))),
        ('enable_min_delta_adc',
            (_compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1361,1362))),
        ('threshold_polarity',
            (_compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1362,1363))),
        ('reset_length',
            (_compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int), 0, 7), (1363,1366))),
        ('mark_first_packet',
            (_compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1366,1367))),
        ('reset_threshold',
            (_basic_property, (int, 0, 255), (1368,1376))),
        ('min_delta_adc',
            (_basic_property, (int, 0, 255), (1376,1384))),
        ('digital_threshold',
            (_list_property, (int, 0, 255, Configuration_v2.num_channels, 8), (1384,1896))),
    ])

# GENERATE THE PROPERTIES!
Configuration_v2.bit_map = OrderedDict()
Configuration_v2.register_map = OrderedDict()
Configuration_v2.register_names = []

for _name, _prop_config in _property_configuration.items():
    _prop_formula = _prop_config[0]
    _formula_args = _prop_config[1]
    _bit_range = _prop_config[2]

    # Add to class attributes
    Configuration_v2.register_names += [_name]
    Configuration_v2.bit_map[_name] = _bit_range
    Configuration_v2.register_map[_name] = range(_bit_range[0]//8, max(_bit_range[1]//8,_bit_range[0]//8+1))

    # Create properties
    _prop, _prop_data = _prop_formula(_name, Configuration_v2.register_map[_name],
        range(*Configuration_v2.bit_map[_name]), *_formula_args)
    setattr(Configuration_v2, _name, _prop)
    setattr(Configuration_v2, _name+'_data', _prop_data)

# Create a look up table from register address to names
Configuration_v2.register_map_inv = OrderedDict()
for register_name, register_addr_range in Configuration_v2.register_map.items():
    for register_addr in register_addr_range:
        try:
            Configuration_v2.register_map_inv[register_addr] += [register_name]
        except KeyError:
            Configuration_v2.register_map_inv[register_addr] = [register_name]




import bidict

from larpix import configs

class IO(object):
    '''
    Base class for IO objects that explicitly describes the necessary functions
    required by any IO class implementation. Additional functions are not used
    by the larpix core classes.

    '''
    _valid_config_classes = ['IO']
    default_filepath = 'io/default.json'

    def __init__(self):
        '''
        Declaration of IO object

        :ivar is_listening: flag for ``start_listening`` and ``stop_listening``

        '''
        self.is_listening = False
        self._io_group_table = bidict.bidict()

    def load(self, filepath=None):
        '''
        Loads a specified IO configuration

        :param filepath: path to io configuration file (JSON)

        '''
        if filepath is None:
            filepath = self.default_filepath
        config = configs.load(filepath, 'io')
        if (config['io_class'] not in self._valid_config_classes):
            raise RuntimeError('Invalid configuration type for {}'.format(type(self).__name__))
        self._io_group_table = bidict.bidict(config['io_group'])

    def encode(self, packets):
        '''
        Encodes a list of packets into a list of IO message objects

        :param packets: ``list`` of larpix ``Packet`` objects to encode into IO messages

        :returns: ``list`` of IO messages

        '''
        pass

    def decode(self, msgs, **kwargs):
        '''
        Decodes a list of IO message objects into respective larpix ``Packet`` objects

        :param msgs: ``list`` of IO messages

        :param kwargs: additional contextual information required to decode messages (implementation specific)

        :returns: ``list`` of larpix ``Packet`` objects

        '''
        pass

    def send(self, packets):
        '''
        Function for sending larpix packet objects

        :param packets: ``list`` of larpix ``Packet`` objects to send via IO

        :returns: ``None``

        '''
        pass


    def start_listening(self):
        '''
        Function for starting read communications on IO

        :returns: ``None``

        '''
        self.is_listening = True


    def stop_listening(self):
        '''
        Function for halting read communications on IO

        :returns: ``None``

        '''
        self.is_listening = False


    def empty_queue(self):
        '''
        Read and remove the current items in the internal queue. The details of
        the queue implementation is left up to the specific IO class.
        Generally returns all packets that have been read since last call to
        ``start_listening`` or ``empty_queue``, whichever was most recent.

        :returns: ``tuple`` of (``list`` of ``Packet`` objects, raw bytestream)

        '''
        pass





'''
A module to assist serial data logging and debugging
'''
from __future__ import absolute_import
import os, time, atexit

from .dataformatter import default_formatter

class DataLogger(object):
    '''Simple low-level serial data logger for debugging purposes'''

    def __init__(self, filename=None, formatter=default_formatter):
        '''Constructor'''
        self.starttime = time.time() # Time since epoch
        if filename is None:
            # Prepare default log path
            logpath = self._default_logpath()
            self.filename = os.path.join(logpath, self._default_logname())
        else:
            # Use provided log file name
            self.filename = filename
        self.formatter = formatter
        self.bytes_written = 0
        self.buffer = bytearray()
        self._buffer_flush = 32000 # flush at 32000 byte buffer size
        self._log_initialized = False
        # Has log been initialized?
        if not self._log_initialized:
            self._initialize_log()
        # Enable, and ensure buffer flushed on python exit
        self._is_enabled = True
        self._exit_func = atexit.register(self.flush)
        
    def _default_logname(self):
        '''Generate a log file name'''
        log_prefix = 'datalog'
        log_specifier = time.strftime('%Y_%m_%d_%H_%M_%S_%Z')
        log_postfix = '.dat'
        return (log_prefix + '_' + log_specifier + '_' + log_postfix)

    def _default_logpath(self):
        '''Generate a log file path'''
        return os.path.join(os.getenv('LP_DATALOG_DIR',os.getcwd()),'datalog')
    
    def _make_file_header(self):
        '''Generate a file header bytearray'''
        return self.formatter.format_block({'block_type':'file',
                                            'starttime':self.starttime})

    def _initialize_log(self):
        '''Initialize log for first write'''
        # Prepend file format test
        self.buffer += self.formatter.file_open_chunk()
        # Prepend file header
        self.buffer += self._make_file_header()
        # Create path, if needed
        logpath = os.path.dirname(self.filename)
        if not os.path.exists(logpath):
            os.makedirs(logpath)
        self._log_initialized = True
        return
        
    def record(self,data_block_desc):
        '''Log a data block'''
        if not self._is_enabled: return
        data_block_desc['block_type'] = 'data'
        #data_block_desc['time'] = time.time()
        self.buffer += self.formatter.format_block(data_block_desc)
        if len(self.buffer) > self._buffer_flush:
            # Flush current data to file
            self.flush()
        return

    def flush(self):
        '''Write current data blocks to file'''
        # Check if enabled
        if not self._is_enabled: return
        # Is there data to log?
        if len(self.buffer) == 0: return  # No data
        # Write data to file
        with open(self.filename,'ab') as log_file:
            log_file.write(self.buffer)
        # Track total bytes written
        self.bytes_written += len(self.buffer)
        # Clear buffer
        self.buffer = bytearray()
        return

    def enable(self):
        '''Enable logger'''
        self._is_enabled = True
        return
    
    def disable(self):
        '''Flush logger and disable'''
        self.flush()
        self._is_enabled = False
        return

    def is_enabled(self):
        '''Return true if logger is enabled'''
        return self._is_enabled



from bitarray import bitarray
import os
import errno

from .. import bitarrayhelper as bah
from .. import configs
from . import BaseConfiguration, _Smart_List

class Configuration_v1(BaseConfiguration):
    '''
    Represents the desired configuration state of a LArPix v1 chip.

    '''

    asic_version = 1
    default_configuration_file = 'chip/default.json'
    num_registers = 63
    num_channels = 32
    pixel_trim_threshold_addresses = list(range(0, 32))
    global_threshold_address = 32
    csa_gain_and_bypasses_address = 33
    csa_bypass_select_addresses = list(range(34, 38))
    csa_monitor_select_addresses = list(range(38, 42))
    csa_testpulse_enable_addresses = list(range(42, 46))
    csa_testpulse_dac_amplitude_address = 46
    test_mode_xtrig_reset_diag_address = 47
    sample_cycles_address = 48
    test_burst_length_addresses = [49, 50]
    adc_burst_length_address = 51
    channel_mask_addresses = list(range(52, 56))
    external_trigger_mask_addresses = list(range(56, 60))
    reset_cycles_addresses = [60, 61, 62]
    register_names = ['pixel_trim_thresholds',
                           'global_threshold',
                           'csa_gain',
                           'csa_bypass',
                           'internal_bypass',
                           'csa_bypass_select',
                           'csa_monitor_select',
                           'csa_testpulse_enable',
                           'csa_testpulse_dac_amplitude',
                           'test_mode',
                           'cross_trigger_mode',
                           'periodic_reset',
                           'fifo_diagnostic',
                           'sample_cycles',
                           'test_burst_length',
                           'adc_burst_length',
                           'channel_mask',
                           'external_trigger_mask',
                           'reset_cycles']
    '''
    This attribute lists the names of all available configuration
    registers. Each register name is available as its own attribute for
    inspecting and setting the value of the corresponding register.

    Certain configuration values are set channel-by-channel. These are
    represented by a list of values. For example:

        >>> conf.pixel_trim_thresholds[2:5]
        [16, 16, 16]
        >>> conf.channel_mask[20] = 1
        >>> conf.external_trigger_mask = [0] * 32

    Additionally, other configuration values take up more than or less
    than one complete register. These are still set by referencing the
    appropriate name. For example, ``cross_trigger_mode`` shares a
    register with a few other values, and adjusting the value of the
    ``cross_trigger_mode`` attribute will leave the other values
    unchanged.

    '''

    TEST_OFF = 0x0
    TEST_UART = 0x1
    TEST_FIFO = 0x2
    def __init__(self):
        # Actual setup
        super(Configuration_v1, self).__init__()

        # Annoying things we have to do because the configuration
        # register follows complex semantics:
        # The following dicts/lists specify how to translate a register
        # address into a sensible update to the Configuration object.
        # Simple registers are just the value stored in the register.
        self._simple_registers = {
                32: 'global_threshold',
                46: 'csa_testpulse_dac_amplitude',
                48: 'sample_cycles',
                51: 'adc_burst_length',
                }
        # These registers need the attribute extracted from the register
        # data.
        self._complex_modify_data = {
                33: [('csa_gain', lambda data:data % 2),
                     ('csa_bypass', lambda data:(data//2) % 2),
                     ('internal_bypass', lambda data:(data//8) % 2)],
                47: [('test_mode', lambda data:data % 4),
                     ('cross_trigger_mode', lambda data:(data//4) % 2),
                     ('periodic_reset', lambda data:(data//8) % 2),
                     ('fifo_diagnostic', lambda data:(data//16) % 2)]
                }
        # These registers combine the register data with the existing
        # attribute value to get the new attribute value.
        self._complex_modify_attr = {
                49: ('test_burst_length', lambda val,data:(val//256)*256+data),
                50: ('test_burst_length', lambda val,data:(val%256)+data*256),
                60: ('reset_cycles', lambda val,data:(val//256)*256+data),
                61: ('reset_cycles',
                    lambda val,data:(val//0x10000)*0x10000+data*256+val%256),
                62: ('reset_cycles', lambda val,data:(val%0x10000)+data*0x10000)
                }
        # These registers store 32 bits over 4 registers each, and those
        # 32 bits correspond to entries in a 32-entry list.
        self._complex_array_spec = [
                (range(34, 38), 'csa_bypass_select'),
                (range(38, 42), 'csa_monitor_select'),
                (range(42, 46), 'csa_testpulse_enable'),
                (range(52, 56), 'channel_mask'),
                (range(56, 60), 'external_trigger_mask')]
        self._complex_array = {}
        for addresses, label in self._complex_array_spec:
            for i, address in enumerate(addresses):
                self._complex_array[address] = (label, i)
        # These registers each correspond to an entry in an array
        self._trim_registers = list(range(32))

    @property
    def pixel_trim_thresholds(self):
        return self._pixel_trim_thresholds

    @pixel_trim_thresholds.setter
    def pixel_trim_thresholds(self, values):
        low = 0
        high = 31
        if not (type(values) == list or type(values) == _Smart_List):
            raise ValueError("pixel_trim_threshold is not list")
        if not len(values) == Configuration_v1.num_channels:
            raise ValueError("pixel_trim_threshold length is not %d" % Configuration_v1.num_channels)
        if not all(type(value) == int for value in values):
            raise ValueError("pixel_trim_threshold is not int")
        if any(value > high or value < low for value in values):
            raise ValueError("pixel_trim_threshold out of bounds")

        self._pixel_trim_thresholds = _Smart_List(values, low, high)

    @property
    def global_threshold(self):
        return self._global_threshold

    @global_threshold.setter
    def global_threshold(self, value):
        if not type(value) == int:
            raise ValueError("global_threshold is not int")
        if value > 255 or value < 0:
            raise ValueError("global_threshold out of bounds")

        self._global_threshold = value

    @property
    def csa_gain(self):
        return self._csa_gain

    @csa_gain.setter
    def csa_gain(self, value):
        if not type(value) == int:
            raise ValueError("csa_gain is not int")
        if value > 1 or value < 0:
            raise ValueError("csa_gain out of bounds")

        self._csa_gain = value

    @property
    def csa_bypass(self):
        return self._csa_bypass

    @csa_bypass.setter
    def csa_bypass(self, value):
        if not type(value) == int:
            raise ValueError("csa_bypass is not int")
        if value > 1 or value < 0:
            raise ValueError("csa_bypass out of bounds")

        self._csa_bypass = value

    @property
    def internal_bypass(self):
        return self._internal_bypass

    @internal_bypass.setter
    def internal_bypass(self, value):
        if not type(value) == int:
            raise ValueError("internal_bypass is not int")
        if value > 1 or value < 0:
            raise ValueError("internal_bypass out of bounds")

        self._internal_bypass = value

    @property
    def csa_bypass_select(self):
        return self._csa_bypass_select

    @csa_bypass_select.setter
    def csa_bypass_select(self, values):
        low = 0
        high = 1
        if not (type(values) == list or type(values) == _Smart_List):
            raise ValueError("csa_bypass_select is not list")
        if not len(values) == Configuration_v1.num_channels:
            raise ValueError("csa_bypass_select length is not %d" % Configuration_v1.num_channels)
        if not all(type(value) == int for value in values):
            raise ValueError("csa_bypass_select is not int")
        if any(value > high or value < low for value in values):
            raise ValueError("csa_bypass_select out of bounds")

        self._csa_bypass_select = _Smart_List(values, low, high)

    @property
    def csa_monitor_select(self):
        return self._csa_monitor_select

    @csa_monitor_select.setter
    def csa_monitor_select(self, values):
        low = 0
        high = 1
        if not (type(values) == list or type(values) == _Smart_List):
            raise ValueError("csa_monitor_select is not list")
        if not len(values) == Configuration_v1.num_channels:
            raise ValueError("csa_monitor_select length is not %d" % Configuration_v1.num_channels)
        if not all(type(value) == int for value in values):
            raise ValueError("csa_monitor_select is not int")
        if any(value > high or value < low for value in values):
            raise ValueError("csa_monitor_select out of bounds")

        self._csa_monitor_select = _Smart_List(values, low, high)

    @property
    def csa_testpulse_enable(self):
        return self._csa_testpulse_enable

    @csa_testpulse_enable.setter
    def csa_testpulse_enable(self, values):
        if not type(values) == list:
            raise ValueError("csa_testpulse_enable is not list")
        if not len(values) == Configuration_v1.num_channels:
            raise ValueError("csa_testpulse_enable length is not %d" % Configuration_v1.num_channels)
        if not all(type(value) == int for value in values):
            raise ValueError("csa_testpulse_enable is not int")
        if any(value > 1 or value < 0 for value in values):
            raise ValueError("csa_testpulse_enable out of bounds")

        self._csa_testpulse_enable = values

    @property
    def csa_testpulse_dac_amplitude(self):
        return self._csa_testpulse_dac_amplitude

    @csa_testpulse_dac_amplitude.setter
    def csa_testpulse_dac_amplitude(self, value):
        if not type(value) == int:
            raise ValueError("csa_testpulse_dac_amplitude is not int")
        if value > 255 or value < 0:
            raise ValueError("csa_testpulse_dac_amplitude out of bounds")

        self._csa_testpulse_dac_amplitude = value

    @property
    def test_mode(self):
        return self._test_mode

    @test_mode.setter
    def test_mode(self, value):
        if not type(value) == int:
            raise ValueError("test_mode is not int")
        valid_values = [Configuration_v1.TEST_OFF, Configuration_v1.TEST_UART,
                        Configuration_v1.TEST_FIFO]
        if not value in valid_values:
            raise ValueError("test_mode is not valid")

        self._test_mode = value

    @property
    def cross_trigger_mode(self):
        return self._cross_trigger_mode

    @cross_trigger_mode.setter
    def cross_trigger_mode(self, value):
        if not type(value) == int:
            raise ValueError("cross_trigger_mode is not int")
        if value > 1 or value < 0:
            raise ValueError("cross_trigger_mode out of bounds")

        self._cross_trigger_mode = value

    @property
    def periodic_reset(self):
        return self._periodic_reset

    @periodic_reset.setter
    def periodic_reset(self, value):
        if not type(value) == int:
            raise ValueError("periodic_reset is not int")
        if value > 1 or value < 0:
            raise ValueError("periodic_reset out of bounds")

        self._periodic_reset = value

    @property
    def fifo_diagnostic(self):
        return self._fifo_diagnostic

    @fifo_diagnostic.setter
    def fifo_diagnostic(self, value):
        if not type(value) == int:
            raise ValueError("fifo_diagnostic is not int")
        if value > 1 or value < 0:
            raise ValueError("fifo_diagnostic out of bounds")

        self._fifo_diagnostic = value

    @property
    def sample_cycles(self):
        return self._sample_cycles

    @sample_cycles.setter
    def sample_cycles(self, value):
        if not type(value) == int:
            raise ValueError("sample_cycles is not int")
        if value > 255 or value < 0:
            raise ValueError("sample_cycles out of bounds")

        self._sample_cycles = value

    @property
    def test_burst_length(self):
        return self._test_burst_length

    @test_burst_length.setter
    def test_burst_length(self, value):
        if not type(value) == int:
            raise ValueError("test_burst_length is not int")
        if value > 65535 or value < 0:
            raise ValueError("test_burst_length out of bounds")

        self._test_burst_length = value

    @property
    def adc_burst_length(self):
        return self._adc_burst_length

    @adc_burst_length.setter
    def adc_burst_length(self, value):
        if not type(value) == int:
            raise ValueError("adc_burst_length is not int")
        if value > 255 or value < 0:
            raise ValueError("adc_burst_length out of bounds")

        self._adc_burst_length = value

    @property
    def channel_mask(self):
        return self._channel_mask

    @channel_mask.setter
    def channel_mask(self, values):
        if not type(values) == list:
            raise ValueError("channel_mask is not list")
        if not len(values) == Configuration_v1.num_channels:
            raise ValueError("channel_mask length is not %d" % Configuration_v1.num_channels)
        if not all(type(value) == int for value in values):
            raise ValueError("channel_mask is not int")
        if any(value > 1 or value < 0 for value in values):
            raise ValueError("channel_mask out of bounds")

        self._channel_mask = values

    @property
    def external_trigger_mask(self):
        return self._external_trigger_mask

    @external_trigger_mask.setter
    def external_trigger_mask(self, values):
        if not type(values) == list:
            raise ValueError("external_trigger_mask is not list")
        if not len(values) == Configuration_v1.num_channels:
            raise ValueError("external_trigger_mask length is not %d" % Configuration_v1.num_channels)
        if not all(type(value) == int for value in values):
            raise ValueError("external_trigger_mask is not int")
        if any(value > 1 or value < 0 for value in values):
            raise ValueError("external_trigger_mask out of bounds")

        self._external_trigger_mask = values

    @property
    def reset_cycles(self):
        return self._reset_cycles

    @reset_cycles.setter
    def reset_cycles(self, value):
        if not type(value) == int:
            raise ValueError("reset_cycles is not int")
        if value > 16777215 or value < 0:
            raise ValueError("reset_cycles out of bounds")

        self._reset_cycles = value

    def enable_channels(self, list_of_channels=None):
        '''
        Shortcut for changing the channel mask for the given
        channels to "enable" (i.e. 0).

        '''
        if list_of_channels is None:
            list_of_channels = range(Configuration_v1.num_channels)
        for channel in list_of_channels:
            self.channel_mask[channel] = 0

    def disable_channels(self, list_of_channels=None):
        '''
        Shortcut for changing the channel mask for the given channels
        to "disable" (i.e. 1).

        '''
        if list_of_channels is None:
            list_of_channels = range(Configuration_v1.num_channels)
        for channel in list_of_channels:
            self.channel_mask[channel] = 1

    def enable_external_trigger(self, list_of_channels=None):
        '''
        Shortcut for enabling the external trigger functionality for the
        given channels. (I.e. disabling the mask.)

        '''
        if list_of_channels is None:
            list_of_channels = range(Configuration_v1.num_channels)
        for channel in list_of_channels:
            self.external_trigger_mask[channel] = 0

    def disable_external_trigger(self, list_of_channels=None):
        '''
        Shortcut for disabling the external trigger functionality for
        the given channels. (I.e. enabling the mask.)

        '''
        if list_of_channels is None:
            list_of_channels = range(Configuration_v1.num_channels)
        for channel in list_of_channels:
            self.external_trigger_mask[channel] = 1

    def enable_testpulse(self, list_of_channels=None):
        '''
        Shortcut for enabling the test pulser for the given channels.

        '''
        if list_of_channels is None:
            list_of_channels = range(Configuration_v1.num_channels)
        for channel in list_of_channels:
            self.csa_testpulse_enable[channel] = 0

    def disable_testpulse(self, list_of_channels=None):
        '''
        Shortcut for disabling the test pulser for the given channels.

        '''
        if list_of_channels is None:
            list_of_channels = range(Configuration_v1.num_channels)
        for channel in list_of_channels:
            self.csa_testpulse_enable[channel] = 1

    def enable_analog_monitor(self, channel):
        '''
        Shortcut for enabling the analog monitor on the given channel.

        '''
        self.csa_monitor_select[channel] = 1

    def disable_analog_monitor(self):
        '''
        Shortcut for disabling the analog monitor (on all channels).

        '''
        self.csa_monitor_select = [0] * Configuration_v1.num_channels

    def all_data(self, **kwargs):
        bits = []
        num_channels = Configuration_v1.num_channels
        for channel in range(num_channels):
            bits.append(self.trim_threshold_data(channel))
        bits.append(self.global_threshold_data())
        bits.append(self.csa_gain_and_bypasses_data())
        for chunk in range(4):
            bits.append(self.csa_bypass_select_data(chunk))
        for chunk in range(4):
            bits.append(self.csa_monitor_select_data(chunk))
        for chunk in range(4):
            bits.append(self.csa_testpulse_enable_data(chunk))
        bits.append(self.csa_testpulse_dac_amplitude_data())
        bits.append(self.test_mode_xtrig_reset_diag_data())
        bits.append(self.sample_cycles_data())
        bits.append(self.test_burst_length_data(0))
        bits.append(self.test_burst_length_data(1))
        bits.append(self.adc_burst_length_data())
        for chunk in range(4):
            bits.append(self.channel_mask_data(chunk))
        for chunk in range(4):
            bits.append(self.external_trigger_mask_data(chunk))
        bits.append(self.reset_cycles_data(0))
        bits.append(self.reset_cycles_data(1))
        bits.append(self.reset_cycles_data(2))
        return bits

    def from_dict_registers(self, d, **kwargs):
        '''
        Load in the configuration specified by a dict of (register,
        value) pairs.

        '''
        def bits_to_array(data):
            bits = bah.fromuint(data, 8)
            return [int(bit) for bit in bits][::-1]

        for address, value in d.items():
            if address in self._simple_registers:
                setattr(self, self._simple_registers[address], value)
            elif address in self._complex_modify_data:
                attributes = self._complex_modify_data[address]
                for name, extract in attributes:
                    setattr(self, name, extract(value))
            elif address in self._complex_modify_attr:
                name, combine = self._complex_modify_attr[address]
                current_value = getattr(self, name)
                setattr(self, name, combine(current_value, value))
            elif address in self._complex_array:
                name, index = self._complex_array[address]
                affected = slice(index*8, (index+1)*8)
                attr_list = getattr(self, name)
                attr_list[affected] = bits_to_array(value)
            elif address in self._trim_registers:
                self.pixel_trim_thresholds[address] = value
        return  #phew

    def trim_threshold_data(self, channel):
        return bah.fromuint(self.pixel_trim_thresholds[channel], 8)

    def global_threshold_data(self):
        return bah.fromuint(self.global_threshold, 8)

    def csa_gain_and_bypasses_data(self):
        return bitarray('0000') + [self.internal_bypass, 0,
                self.csa_bypass, self.csa_gain]

    def csa_bypass_select_data(self, chunk):
        if chunk == 0:
            return bitarray(self.csa_bypass_select[7::-1])
        else:
            high_bit = (chunk + 1) * 8 - 1
            low_bit = chunk * 8 - 1
            return bitarray(self.csa_bypass_select[high_bit:low_bit:-1])

    def csa_monitor_select_data(self, chunk):
        if chunk == 0:
            return bitarray(self.csa_monitor_select[7::-1])
        else:
            high_bit = (chunk + 1) * 8 - 1
            low_bit = chunk * 8 - 1
            return bitarray(self.csa_monitor_select[high_bit:low_bit:-1])

    def csa_testpulse_enable_data(self, chunk):
        if chunk == 0:
            return bitarray(self.csa_testpulse_enable[7::-1])
        else:
            high_bit = (chunk + 1) * 8 - 1
            low_bit = chunk * 8 - 1
            return bitarray(self.csa_testpulse_enable[high_bit:low_bit:-1])

    def csa_testpulse_dac_amplitude_data(self):
        return bah.fromuint(self.csa_testpulse_dac_amplitude, 8)

    def test_mode_xtrig_reset_diag_data(self):
        toReturn = bitarray([0, 0, 0, self.fifo_diagnostic,
            self.periodic_reset,
            self.cross_trigger_mode])
        toReturn.extend(bah.fromuint(self.test_mode, 2))
        return toReturn

    def sample_cycles_data(self):
        return bah.fromuint(self.sample_cycles, 8)

    def test_burst_length_data(self, chunk):
        bits = bah.fromuint(self.test_burst_length, 16)
        if chunk == 0:
            return bits[8:]
        elif chunk == 1:
            return bits[:8]

    def adc_burst_length_data(self):
        return bah.fromuint(self.adc_burst_length, 8)

    def channel_mask_data(self, chunk):
        if chunk == 0:
            return bitarray(self.channel_mask[7::-1])
        else:
            high_bit = (chunk + 1) * 8 - 1
            low_bit = chunk * 8 - 1
            return bitarray(self.channel_mask[high_bit:low_bit:-1])

    def external_trigger_mask_data(self, chunk):
        if chunk == 0:
            return bitarray(self.external_trigger_mask[7::-1])
        else:
            high_bit = (chunk + 1) * 8 - 1
            low_bit = chunk * 8 - 1
            return bitarray(self.external_trigger_mask[high_bit:low_bit:-1])

    def reset_cycles_data(self, chunk):
        bits = bah.fromuint(self.reset_cycles, 24)
        if chunk == 0:
            return bits[16:]
        elif chunk == 1:
            return bits[8:16]
        elif chunk == 2:
            return bits[:8]
